from abc import ABCMeta

import six

from dagster import check
from dagster.core.types.field_utils import check_user_facing_opt_field_param
from dagster.utils import frozendict

from .input import InputDefinition
from .output import OutputDefinition
from .utils import check_valid_name


class ISolidDefinition(six.with_metaclass(ABCMeta)):
    def __init__(self, name, input_dict, output_dict, description=None, metadata=None):
        self.name = check_valid_name(name)
        self.description = check.opt_str_param(description, 'description')
        self.metadata = check.opt_dict_param(metadata, 'metadata', key_type=str)
        self.input_dict = frozendict(input_dict)
        self.output_dict = frozendict(output_dict)

    def has_input(self, name):
        check.str_param(name, 'name')
        return name in self.input_dict

    def input_def_named(self, name):
        check.str_param(name, 'name')
        return self.input_dict[name]

    def has_output(self, name):
        check.str_param(name, 'name')
        return name in self.output_dict

    def output_def_named(self, name):
        check.str_param(name, 'name')
        return self.output_dict[name]


class SolidDefinition(ISolidDefinition):
    '''A solid (a name extracted from the acronym of "software-structured data" (SSD)) represents
    a unit of computation within a data pipeline.

    As its core, a solid is a function. It accepts inputs (which are values produced from
    other solids) and configuration, and produces outputs. These solids are composed as a
    directed, acyclic graph (DAG) within a pipeline to form a computation that produces
    data assets.

    Solids should be implemented as idempotent, parameterizable, non-destructive functions.
    Data computations with these properties are much easier to test, reason about, and operate.

    The inputs and outputs are gradually, optionally typed by the dagster type system. Types
    can be user-defined and can represent entities as varied as scalars, dataframe, database
    tables, and so forth. They can represent pure in-memory objects, or handles to assets
    on disk or in external resources.

    A solid is a generalized abstraction that could take many forms.

    End users should prefer the @solid and @lambda_solid decorator. SolidDefinition
    is generally used by framework authors.

    Args:
        name (str): Name of the solid.
        inputs (List[InputDefinition]): Inputs of the solid.

        transform_fn (Callable[[SystemTransformExecutionContext, ], Iterable[Union[Result,
            Materialization]]]): The core of the solid, the function that does the actual
            computation. The arguments passed to this function after context are deteremined by
            ``inputs``.

            This function yields :py:class:`Result` according to ``outputs`` or
            :py:class:`Materialization`.

        outputs (List[OutputDefinition]): Outputs of the solid.
        config_field (Optional[Field]): How the solid configured.
        description (Optional[str]): Description of the solid.
        metadata (Optional[Dict[Any, Any]]):
            Arbitrary metadata for the solid. Some frameworks expect and require
            certain metadata to be attached to a solid.
        resources (Optional[Set[str]]): List of resources handles required by this solid.

    Examples:
        .. code-block:: python

            def _add_one(_context, inputs):
                yield Result(inputs["num"] + 1)

            SolidDefinition(
                name="add_one",
                inputs=[InputDefinition("num", Int)],
                outputs=[OutputDefinition(Int)], # default name ("result")
                transform_fn=_add_one,
            )
    '''

    def __init__(
        self,
        name,
        inputs,
        transform_fn,
        outputs,
        config_field=None,
        description=None,
        metadata=None,
        resources=None,
        step_metadata_fn=None,
    ):
        self.name = check_valid_name(name)
        self.transform_fn = check.callable_param(transform_fn, 'transform_fn')
        self.description = check.opt_str_param(description, 'description')
        self.config_field = check_user_facing_opt_field_param(
            config_field,
            'config_field',
            'of a SolidDefinition or @solid named "{name}"'.format(name=name),
        )
        self.metadata = check.opt_dict_param(metadata, 'metadata', key_type=str)
        self.resources = check.opt_set_param(resources, 'resources', of_type=str)
        self.step_metadata_fn = step_metadata_fn

        input_dict = {inp.name: inp for inp in check.list_param(inputs, 'inputs', InputDefinition)}
        output_dict = {
            output.name: output for output in check.list_param(outputs, 'outputs', OutputDefinition)
        }

        super(SolidDefinition, self).__init__(name, input_dict, output_dict, description, metadata)


class CompositeSolidDefinition(ISolidDefinition):
    def __init__(self, name, description=None, metadata=None):

        super(CompositeSolidDefinition, self).__init__(name, {}, {}, description, metadata)



from dagster import (
    DagsterEvaluateConfigValueError,
    DagsterInvariantViolationError,
    ExecutionContext,
    InProcessExecutorConfig,
    PipelineDefinition,
    PipelineContextDefinition,
    Result,
    RunConfig,
    SolidDefinition,
    check,
    execute_pipeline,
)

from dagster.core.types.evaluator import evaluate_config_value


def execute_single_solid_in_isolation(
    context_params, solid_def, environment=None, raise_on_error=True
):
    '''
    Deprecated.

    Execute a solid outside the context of a pipeline, with an already-created context.

    Prefer execute_solid in dagster.utils.test
    '''
    check.inst_param(context_params, 'context_params', ExecutionContext)
    check.inst_param(solid_def, 'solid_def', SolidDefinition)
    environment = check.opt_dict_param(environment, 'environment')
    check.bool_param(raise_on_error, 'raise_on_error')

    single_solid_environment = {
        'expectations': environment.get('expectations'),
        'context': environment.get('context'),
        'solids': {solid_def.name: environment['solids'][solid_def.name]}
        if solid_def.name in environment.get('solids', {})
        else None,
    }

    pipeline_result = execute_pipeline(
        PipelineDefinition(
            solids=[solid_def],
            context_definitions=PipelineContextDefinition.passthrough_context_definition(
                context_params
            ),
        ),
        environment_dict=single_solid_environment,
        run_config=RunConfig(
            executor_config=InProcessExecutorConfig(raise_on_error=raise_on_error)
        ),
    )

    return pipeline_result


def single_output_transform(name, inputs, transform_fn, output, description=None):
    '''It is commmon to want a Solid that has only inputs, a single output (with the default
    name), and no config. So this is a helper function to do that. This transform function
    must return the naked return value (as opposed to a Result object).

    Args:
        name (str): Name of the solid.
        inputs (List[InputDefinition]): Inputs of solid.
        transform_fn (callable):
            Callable with the signature
            (context: ExecutionContext, inputs: Dict[str, Any]) : Any
        output (OutputDefinition): Output of the solid.
        description (str): Descripion of the solid.

    Returns:
        SolidDefinition:

    Example:

        .. code-block:: python

            single_output_transform(
                'add_one',
                inputs=InputDefinition('num', types.Int),
                output=OutputDefinition(types.Int),
                transform_fn=lambda context, inputs: inputs['num'] + 1
            )

    '''

    def _new_transform_fn(context, inputs):
        value = transform_fn(context, inputs)
        if isinstance(value, Result):
            raise DagsterInvariantViolationError(
                '''Single output transform Solid {name} returned a Result. Just return
                value directly without wrapping it in Result'''
            )
        yield Result(value=value)

    return SolidDefinition(
        name=name,
        inputs=inputs,
        transform_fn=_new_transform_fn,
        outputs=[output],
        description=description,
    )


# This is a legacy API from when the config parsing only returned a single
# error. Existing test logic was written assuming structure to this is still
# around to avoid having to port all the unit tests.
def throwing_evaluate_config_value(config_type, config_value):
    result = evaluate_config_value(config_type, config_value)
    if not result.success:
        raise DagsterEvaluateConfigValueError(result.errors[0].stack, result.errors[0].message)
    return result.value



'''
This module contains the execution context objects that are internal to the system.
Not every property on these should be exposed to random Jane or Joe dagster user
so we have a different layer of objects that encode the explicit public API
in the user_context module
'''
from collections import namedtuple
import uuid
import multiprocessing

from dagster import check
from dagster.utils import merge_dicts
from dagster.core.errors import DagsterInvariantViolationError

from dagster.core.definitions.expectation import ExpectationDefinition
from dagster.core.definitions.input import InputDefinition
from dagster.core.definitions.output import OutputDefinition
from dagster.core.log import DagsterLog
from dagster.core.runs import RunStorageMode, RunStorage
from dagster.core.system_config.objects import EnvironmentConfig


class ExecutorConfig:
    pass


class InProcessExecutorConfig(ExecutorConfig):
    def __init__(self, raise_on_error=True):
        self.raise_on_error = check.bool_param(raise_on_error, 'raise_on_error')


class MultiprocessExecutorConfig(ExecutorConfig):
    def __init__(self, pipeline_fn, max_concurrent=None):
        self.pipeline_fn = check.callable_param(pipeline_fn, 'pipeline_fn')
        max_concurrent = (
            max_concurrent if max_concurrent is not None else multiprocessing.cpu_count()
        )
        self.max_concurrent = check.int_param(max_concurrent, 'max_concurrent')
        check.invariant(self.max_concurrent > 0, 'max_concurrent processes must be greater than 0')
        self.raise_on_error = False


def make_new_run_id():
    return str(uuid.uuid4())


class ReexecutionConfig:
    def __init__(self, previous_run_id, step_output_handles):
        self.previous_run_id = previous_run_id
        self.step_output_handles = step_output_handles


class RunConfig(
    namedtuple(
        '_RunConfig',
        (
            'run_id tags event_callback loggers executor_config storage_mode reexecution_config '
            'step_keys_to_execute mode'
        ),
    )
):
    '''
    Configuration that controls the details of how Dagster will execute a pipeline.

    Args:
      run_id (str): The ID to use for this run. If not provided a new UUID will
        be created using `uuid4`.
      tags (dict[str, str]): Key value pairs that will be added to logs.
      event_callback (callable): A callback to invoke with each :py:class:`EventRecord`
        produced during execution.
      loggers (list): Additional loggers that log messages will be sent to.
      executor_config (ExecutorConfig): Configuration for where and how computation will occur.
      storage_mode (RunStorageMode): Where intermediate artifacts will be stored during execution.
      rexecution_config (RexecutionConfig): Information about a previous run to allow
        for subset rexecution.
      step_keys_to_execute (list[str]): They subset of steps from a pipeline to execute this run.
    '''

    def __new__(
        cls,
        run_id=None,
        tags=None,
        event_callback=None,
        loggers=None,
        executor_config=None,
        storage_mode=None,
        reexecution_config=None,
        step_keys_to_execute=None,
        mode=None,
    ):
        if (
            isinstance(executor_config, MultiprocessExecutorConfig)
            and storage_mode is RunStorageMode.IN_MEMORY
        ):
            raise DagsterInvariantViolationError(
                'Can not create a RunConfig with executor_config MultiProcessExecutorConfig and '
                'storage_mode RunStorageMode.IN_MEMORY'
            )

        check.opt_list_param(step_keys_to_execute, 'step_keys_to_execute', of_type=str)

        return super(RunConfig, cls).__new__(
            cls,
            run_id=check.str_param(run_id, 'run_id') if run_id else make_new_run_id(),
            tags=check.opt_dict_param(tags, 'tags', key_type=str, value_type=str),
            event_callback=check.opt_callable_param(event_callback, 'event_callback'),
            loggers=check.opt_list_param(loggers, 'loggers'),
            executor_config=check.inst_param(executor_config, 'executor_config', ExecutorConfig)
            if executor_config
            else InProcessExecutorConfig(),
            storage_mode=check.opt_inst_param(storage_mode, 'storage_mode', RunStorageMode),
            reexecution_config=check.opt_inst_param(
                reexecution_config, 'reexecution_config', ReexecutionConfig
            ),
            step_keys_to_execute=step_keys_to_execute,
            mode=check.opt_str_param(mode, 'mode'),
        )

    @staticmethod
    def nonthrowing_in_process():
        return RunConfig(executor_config=InProcessExecutorConfig(raise_on_error=False))

    def with_tags(self, **new_tags):
        new_tags = merge_dicts(self.tags, new_tags)
        return RunConfig(**merge_dicts(self._asdict(), {'tags': new_tags}))


class SystemPipelineExecutionContextData(
    namedtuple(
        '_SystemPipelineExecutionContextData',
        (
            'run_config resources environment_config pipeline_def '
            'run_storage intermediates_manager'
        ),
    )
):
    '''
    PipelineContextData is the data that remains context throughtout the entire execution
    of a pipeline.
    '''

    def __new__(
        cls,
        run_config,
        resources,
        environment_config,
        pipeline_def,
        run_storage,
        intermediates_manager,
    ):
        from dagster.core.definitions import PipelineDefinition
        from dagster.core.intermediates_manager import IntermediatesManager

        return super(SystemPipelineExecutionContextData, cls).__new__(
            cls,
            run_config=check.inst_param(run_config, 'run_config', RunConfig),
            resources=resources,
            environment_config=check.inst_param(
                environment_config, 'environment_config', EnvironmentConfig
            ),
            pipeline_def=check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition),
            run_storage=check.inst_param(run_storage, 'run_storage', RunStorage),
            intermediates_manager=check.inst_param(
                intermediates_manager, 'intermediates_manager', IntermediatesManager
            ),
        )

    @property
    def run_id(self):
        return self.run_config.run_id

    @property
    def event_callback(self):
        return self.run_config.event_callback

    @property
    def environment_dict(self):
        return self.environment_config.original_config_dict


class SystemPipelineExecutionContext(object):
    __slots__ = ['_pipeline_context_data', '_logging_tags', '_log', '_legacy_context', '_events']

    def __init__(self, pipeline_context_data, logging_tags, log):
        self._pipeline_context_data = check.inst_param(
            pipeline_context_data, 'pipeline_context_data', SystemPipelineExecutionContextData
        )
        self._logging_tags = check.dict_param(logging_tags, 'logging_tags')
        self._log = check.inst_param(log, 'log', DagsterLog)

    def for_step(self, step):
        from dagster.core.execution_plan.objects import ExecutionStep

        check.inst_param(step, 'step', ExecutionStep)

        logging_tags = merge_dicts(self.logging_tags, step.logging_tags)
        log = DagsterLog(self.run_id, logging_tags, self.log.loggers)
        return SystemStepExecutionContext(self._pipeline_context_data, logging_tags, log, step)

    @property
    def executor_config(self):
        return self.run_config.executor_config

    @property
    def run_config(self):
        return self._pipeline_context_data.run_config

    @property
    def resources(self):
        return self._pipeline_context_data.resources.build()

    @property
    def run_id(self):
        return self._pipeline_context_data.run_id

    @property
    def environment_dict(self):
        return self._pipeline_context_data.environment_dict

    @property
    def environment_config(self):
        return self._pipeline_context_data.environment_config

    @property
    def logging_tags(self):
        return self._logging_tags

    def has_tag(self, key):
        check.str_param(key, 'key')
        return key in self._logging_tags

    def get_tag(self, key):
        check.str_param(key, 'key')
        return self._logging_tags[key]

    @property
    def pipeline_def(self):
        return self._pipeline_context_data.pipeline_def

    @property
    def event_callback(self):
        return self._pipeline_context_data.event_callback

    def has_event_callback(self):
        return self._pipeline_context_data.event_callback is not None

    @property
    def log(self):
        return self._log

    @property
    def run_storage(self):
        return self._pipeline_context_data.run_storage

    @property
    def intermediates_manager(self):
        return self._pipeline_context_data.intermediates_manager


class SystemStepExecutionContext(SystemPipelineExecutionContext):
    __slots__ = ['_step']

    def __init__(self, pipeline_context_data, tags, log, step):
        from dagster.core.execution_plan.objects import ExecutionStep

        self._step = check.inst_param(step, 'step', ExecutionStep)
        super(SystemStepExecutionContext, self).__init__(pipeline_context_data, tags, log)

    def for_transform(self):
        return SystemTransformExecutionContext(
            self._pipeline_context_data, self.logging_tags, self.log, self.step
        )

    def for_expectation(self, inout_def, expectation_def):
        return SystemExpectationExecutionContext(
            self._pipeline_context_data,
            self.logging_tags,
            self.log,
            self.step,
            inout_def,
            expectation_def,
        )

    @property
    def step(self):
        return self._step

    @property
    def solid_handle(self):
        return self._step.solid_handle

    @property
    def solid_def(self):
        return self.solid.definition

    @property
    def solid(self):
        return self.pipeline_def.get_solid(self._step.solid_handle)

    @property
    def resources(self):
        return self._pipeline_context_data.resources.build(
            self.solid.resource_mapper_fn, self.solid_def.resources
        )


class SystemTransformExecutionContext(SystemStepExecutionContext):
    @property
    def solid_config(self):
        solid_config = self.environment_config.solids.get(str(self.solid_handle))
        return solid_config.config if solid_config else None


class SystemExpectationExecutionContext(SystemStepExecutionContext):
    __slots__ = ['_inout_def', '_expectation_def']

    def __init__(self, pipeline_context_data, tags, log, step, inout_def, expectation_def):
        self._expectation_def = check.inst_param(
            expectation_def, 'expectation_def', ExpectationDefinition
        )
        self._inout_def = check.inst_param(
            inout_def, 'inout_def', (InputDefinition, OutputDefinition)
        )
        super(SystemExpectationExecutionContext, self).__init__(
            pipeline_context_data, tags, log, step
        )

    @property
    def expectation_def(self):
        return self._expectation_def

    @property
    def inout_def(self):
        return self._inout_def



from .context import PipelineContextDefinition

from .dependency import (
    DependencyDefinition,
    Solid,
    SolidHandle,
    SolidInputHandle,
    SolidInstance,
    SolidOutputHandle,
)

from .environment_schema import (
    EnvironmentSchema,
    create_environment_schema,
    create_environment_type,
)

from .expectation import ExpectationDefinition, ExpectationResult

from .input import InputDefinition

from .output import OutputDefinition

from .resource import ResourceDefinition

from .result import Result

from .materialization import Materialization

from .mode import ModeDefinition

from .repository import RepositoryDefinition, PipelinePreset

from .pipeline import PipelineDefinition, solids_in_topological_order

from .pipeline_creation import create_execution_structure

from .solid import SolidDefinition, ISolidDefinition



from collections import namedtuple

from dagster import check

from dagster.core.runs import InMemoryRunStorage, FileSystemRunStorage
from dagster.core.errors import DagsterInvariantViolationError

DEFAULT_CONTEXT_NAME = 'default'


def _default_persistence_config():
    return {'file': {}}


# lifted from https://bit.ly/2HcQAuv
class ContextConfig(namedtuple('_ContextConfig', 'name config resources persistence')):
    def __new__(cls, name=None, config=None, resources=None, persistence=None):
        return super(ContextConfig, cls).__new__(
            cls,
            check.opt_str_param(name, 'name', DEFAULT_CONTEXT_NAME),
            config,
            check.opt_dict_param(resources, 'resources', key_type=str),
            _default_persistence_config()
            if persistence is None
            else check.dict_param(persistence, 'persistence', key_type=str),
        )


class SolidConfig(namedtuple('_SolidConfig', 'config inputs outputs')):
    def __new__(cls, config, inputs=None, outputs=None):
        return super(SolidConfig, cls).__new__(
            cls,
            config,
            check.opt_dict_param(inputs, 'inputs', key_type=str),
            check.opt_list_param(outputs, 'outputs', of_type=dict),
        )


class EnvironmentConfig(
    namedtuple(
        '_EnvironmentConfig',
        'context solids expectations execution storage resources original_config_dict',
    )
):
    def __new__(
        cls,
        solids=None,
        context=None,
        expectations=None,
        execution=None,
        storage=None,
        resources=None,
        original_config_dict=None,
    ):
        check.opt_inst_param(context, 'context', ContextConfig)
        check.opt_inst_param(expectations, 'expectations', ExpectationsConfig)
        check.opt_inst_param(execution, 'execution', ExecutionConfig)
        check.opt_inst_param(storage, 'storage', StorageConfig)
        check.opt_dict_param(original_config_dict, 'original_config_dict')
        check.opt_dict_param(resources, 'resources', key_type=str)

        if expectations is None:
            expectations = ExpectationsConfig(evaluate=True)

        if execution is None:
            execution = ExecutionConfig()

        return super(EnvironmentConfig, cls).__new__(
            cls,
            context=context,
            solids=check.opt_dict_param(solids, 'solids', key_type=str, value_type=SolidConfig),
            expectations=expectations,
            execution=execution,
            storage=storage,
            resources=resources,
            original_config_dict=original_config_dict,
        )


class ExpectationsConfig(namedtuple('_ExpecationsConfig', 'evaluate')):
    def __new__(cls, evaluate):
        return super(ExpectationsConfig, cls).__new__(
            cls, evaluate=check.bool_param(evaluate, 'evaluate')
        )


class ExecutionConfig(namedtuple('_ExecutionConfig', '')):
    def __new__(cls):
        return super(ExecutionConfig, cls).__new__(cls)


class StorageConfig(namedtuple('_FilesConfig', 'storage_mode storage_config')):
    def __new__(cls, storage_mode, storage_config):
        return super(StorageConfig, cls).__new__(
            cls,
            storage_mode=check.opt_str_param(storage_mode, 'storage_mode'),
            storage_config=check.opt_dict_param(storage_config, 'storage_config', key_type=str),
        )

    def construct_run_storage(self):
        if self.storage_mode == 'filesystem':
            return FileSystemRunStorage()
        elif self.storage_mode == 'in_memory':
            return InMemoryRunStorage()
        elif self.storage_mode == 's3':
            # TODO: Revisit whether we want to use S3 run storage
            return FileSystemRunStorage()
        elif self.storage_mode is None:
            return InMemoryRunStorage()
        else:
            raise DagsterInvariantViolationError(
                'Invalid storage specified {}'.format(self.storage_mode)
            )



import subprocess


def check_script(path):
    subprocess.check_output(['python', path])


def check_cli_execute_file_pipeline(path, pipeline_fn_name, env_file=None):
    cli_cmd = ['python', '-m', 'dagster', 'pipeline', 'execute', '-f', path, '-n', pipeline_fn_name]

    if env_file:
        cli_cmd.append('-e')
        cli_cmd.append(env_file)

    try:
        subprocess.check_output(cli_cmd)
    except subprocess.CalledProcessError as cpe:
        print(cpe)
        raise cpe



from abc import ABCMeta, abstractmethod
from collections import namedtuple, OrderedDict
from enum import Enum
import json
import os
import shutil

import six

from dagster import check, seven
from dagster.utils import mkdir_p, list_pull


def base_run_directory():
    return os.path.join(seven.get_system_temp_directory(), 'dagster', 'runs')


def meta_file(base_dir):
    return os.path.join(base_dir, 'runmeta.jsonl')


class DagsterRunMeta(namedtuple('_DagsterRunMeta', 'run_id timestamp pipeline_name')):
    def __new__(cls, run_id, timestamp, pipeline_name):
        return super(DagsterRunMeta, cls).__new__(
            cls,
            check.str_param(run_id, 'run_id'),
            check.float_param(timestamp, 'timestamp'),
            check.str_param(pipeline_name, 'pipeline_name'),
        )


class RunStorage(six.with_metaclass(ABCMeta)):  # pylint: disable=no-init
    @abstractmethod
    def write_dagster_run_meta(self, dagster_run_meta):
        pass

    def has_run(self, run_id):
        check.str_param(run_id, 'run_id')
        return run_id in self.get_run_ids()

    @abstractmethod
    def get_run_ids(self):
        pass

    @abstractmethod
    def get_run_metas(self):
        pass

    @abstractmethod
    def get_run_meta(self, run_id):
        pass

    @abstractmethod
    def nuke(self):
        pass

    @property
    @abstractmethod
    def is_persistent(self):
        pass


class FileSystemRunStorage(RunStorage):
    def __init__(self, base_dir=None):
        self._base_dir = check.opt_str_param(base_dir, 'base_dir', base_run_directory())
        mkdir_p(base_run_directory())
        self._meta_file = meta_file(self._base_dir)

    def write_dagster_run_meta(self, dagster_run_meta):
        check.inst_param(dagster_run_meta, 'dagster_run_meta', DagsterRunMeta)

        run_dir = os.path.join(self._base_dir, dagster_run_meta.run_id)

        mkdir_p(run_dir)

        with open(self._meta_file, 'a+') as ff:
            ff.write(seven.json.dumps(dagster_run_meta._asdict()) + '\n')

    def get_run_ids(self):
        return list_pull(self.get_run_metas(), 'run_id')

    def get_run_meta(self, run_id):
        check.str_param(run_id, 'run_id')
        for run_meta in self.get_run_metas():
            if run_meta.run_id == run_id:
                return run_meta

        return None

    def get_run_metas(self):
        if not os.path.exists(self._meta_file):
            return []

        run_metas = []
        with open(self._meta_file, 'r') as ff:
            line = ff.readline()
            while line:
                raw_run_meta = json.loads(line)
                run_metas.append(
                    DagsterRunMeta(
                        run_id=raw_run_meta['run_id'],
                        timestamp=raw_run_meta['timestamp'],
                        pipeline_name=raw_run_meta['pipeline_name'],
                    )
                )
                line = ff.readline()

        return run_metas

    def nuke(self):
        shutil.rmtree(self._base_dir)

    @property
    def is_persistent(self):
        return True


class InMemoryRunStorage(RunStorage):
    def __init__(self):
        self._run_metas = OrderedDict()

    def write_dagster_run_meta(self, dagster_run_meta):
        check.inst_param(dagster_run_meta, 'dagster_run_meta', DagsterRunMeta)
        self._run_metas[dagster_run_meta.run_id] = dagster_run_meta

    def get_run_ids(self):
        return list_pull(self.get_run_metas(), 'run_id')

    def get_run_metas(self):
        return list(self._run_metas.values())

    def get_run_meta(self, run_id):
        check.str_param(run_id, 'run_id')
        return self._run_metas[run_id]

    def nuke(self):
        self._run_metas = OrderedDict()

    @property
    def is_persistent(self):
        return False


class RunStorageMode(Enum):
    IN_MEMORY = 'IN_MEMORY'
    FILESYSTEM = 'FILESYSTEM'
    S3 = 'S3'



from collections import defaultdict

from dagster.core.types.runtime import ALL_RUNTIME_BUILTINS
from dagster.core.errors import DagsterInvalidDefinitionError


from .dependency import DependencyStructure, SolidInstance, Solid


def _build_pipeline_solid_dict(solids, name_to_aliases, alias_to_solid_instance):
    pipeline_solids = []
    for solid_def in solids:
        uses_of_solid = name_to_aliases.get(solid_def.name, {solid_def.name})

        for alias in uses_of_solid:
            solid_instance = alias_to_solid_instance.get(alias)
            resource_mapper_fn = (
                solid_instance.resource_mapper_fn
                if solid_instance
                else SolidInstance.default_resource_mapper_fn
            )
            pipeline_solids.append(
                Solid(name=alias, definition=solid_def, resource_mapper_fn=resource_mapper_fn)
            )

    return {ps.name: ps for ps in pipeline_solids}


def create_execution_structure(solids, dependencies_dict):
    '''This builder takes the dependencies dictionary specified during creation of the
    PipelineDefinition object and builds (1) the execution structure and (2) a solid dependency
    dictionary.

    For example, for the following dependencies:

    dep_dict = {
            SolidInstance('giver'): {},
            SolidInstance('sleeper', alias='sleeper_1'): {
                'units': DependencyDefinition('giver', 'out_1')
            },
            SolidInstance('sleeper', alias='sleeper_2'): {
                'units': DependencyDefinition('giver', 'out_2')
            },
            SolidInstance('sleeper', alias='sleeper_3'): {
                'units': DependencyDefinition('giver', 'out_3')
            },
            SolidInstance('sleeper', alias='sleeper_4'): {
                'units': DependencyDefinition('giver', 'out_4')
            },
            SolidInstance('total'): {
                'in_1': DependencyDefinition('sleeper_1', 'total'),
                'in_2': DependencyDefinition('sleeper_2', 'total'),
                'in_3': DependencyDefinition('sleeper_3', 'total'),
                'in_4': DependencyDefinition('sleeper_4', 'total'),
            },
        },

    This will create:

    pipeline_solid_dict = {
        'giver': <dagster.core.definitions.dependency.Solid object>,
        'sleeper_1': <dagster.core.definitions.dependency.Solid object>,
        'sleeper_2': <dagster.core.definitions.dependency.Solid object>,
        'sleeper_3': <dagster.core.definitions.dependency.Solid object>,
        'sleeper_4': <dagster.core.definitions.dependency.Solid object>,
        'total': <dagster.core.definitions.dependency.Solid object>
    }

    as well as a dagster.core.definitions.dependency.DependencyStructure object.
    '''
    # Same as dep_dict but with SolidInstance replaced by alias string
    aliased_dependencies_dict = {}

    # Keep track of solid name -> all aliases used and alias -> name
    name_to_aliases = defaultdict(set)
    alias_to_solid_instance = {}
    alias_to_name = {}

    for solid_key, input_dep_dict in dependencies_dict.items():
        # We allow deps of the form dependencies={'foo': DependencyDefition('bar')}
        # Here, we replace 'foo' with SolidInstance('foo')
        if not isinstance(solid_key, SolidInstance):
            solid_key = SolidInstance(solid_key)

        alias = solid_key.alias or solid_key.name

        name_to_aliases[solid_key.name].add(alias)
        alias_to_solid_instance[alias] = solid_key
        alias_to_name[alias] = solid_key.name
        aliased_dependencies_dict[alias] = input_dep_dict

        for dependency in input_dep_dict.values():
            name_to_aliases[dependency.solid].add(dependency.solid)

    pipeline_solid_dict = _build_pipeline_solid_dict(
        solids, name_to_aliases, alias_to_solid_instance
    )

    _validate_dependencies(aliased_dependencies_dict, pipeline_solid_dict, alias_to_name)

    dependency_structure = DependencyStructure.from_definitions(
        pipeline_solid_dict, aliased_dependencies_dict
    )

    return dependency_structure, pipeline_solid_dict


def _validate_dependencies(dependencies, solid_dict, alias_to_name):
    for from_solid, dep_by_input in dependencies.items():
        for from_input, dep in dep_by_input.items():
            if from_solid == dep.solid:
                raise DagsterInvalidDefinitionError(
                    'Circular reference detected in solid {from_solid} input {from_input}.'.format(
                        from_solid=from_solid, from_input=from_input
                    )
                )

            if not from_solid in solid_dict:
                aliased_solid = alias_to_name.get(from_solid)
                if aliased_solid == from_solid:
                    raise DagsterInvalidDefinitionError(
                        'Solid {solid} in dependency dictionary not found in solid list'.format(
                            solid=from_solid
                        )
                    )
                else:
                    raise DagsterInvalidDefinitionError(
                        (
                            'Solid {aliased_solid} (aliased by {from_solid} in dependency '
                            'dictionary) not found in solid list'
                        ).format(aliased_solid=aliased_solid, from_solid=from_solid)
                    )
            if not solid_dict[from_solid].definition.has_input(from_input):
                input_list = solid_dict[from_solid].definition.input_dict.keys()
                raise DagsterInvalidDefinitionError(
                    'Solid "{from_solid}" does not have input "{from_input}". '.format(
                        from_solid=from_solid, from_input=from_input
                    )
                    + 'Input list: {input_list}'.format(input_list=input_list)
                )

            if not dep.solid in solid_dict:
                raise DagsterInvalidDefinitionError(
                    'Solid {dep.solid} in DependencyDefinition not found in solid list'.format(
                        dep=dep
                    )
                )

            if not solid_dict[dep.solid].definition.has_output(dep.output):
                raise DagsterInvalidDefinitionError(
                    'Solid {dep.solid} does not have output {dep.output}'.format(dep=dep)
                )

            input_def = solid_dict[from_solid].definition.input_def_named(from_input)
            output_def = solid_dict[dep.solid].definition.output_def_named(dep.output)

            _validate_input_output_pair(input_def, output_def, from_solid, dep)


def _validate_input_output_pair(input_def, output_def, from_solid, dep):
    # Currently, we opt to be overly permissive with input/output type mismatches.

    # Here we check for the case where no value will be provided where one is expected.
    if output_def.runtime_type.is_nothing and not input_def.runtime_type.is_nothing:
        raise DagsterInvalidDefinitionError(
            (
                'Input "{input_def.name}" to solid "{from_solid}" can not depend on the output '
                '"{output_def.name}" from solid "{dep.solid}". '
                'Input "{input_def.name} expects a value of type {input_def.runtime_type.name} '
                'and output "{output_def.name}" returns type {output_def.runtime_type.name}{extra}.'
            ).format(
                from_solid=from_solid,
                dep=dep,
                output_def=output_def,
                input_def=input_def,
                extra=' (which produces no value)' if output_def.runtime_type.is_nothing else '',
            )
        )


def construct_runtime_type_dictionary(solid_defs):
    type_dict = {t.name: t for t in ALL_RUNTIME_BUILTINS}
    for solid_def in solid_defs:
        for input_def in solid_def.input_dict.values():
            type_dict[input_def.runtime_type.name] = input_def.runtime_type
            for inner_type in input_def.runtime_type.inner_types:
                type_dict[inner_type.name] = inner_type

        for output_def in solid_def.output_dict.values():
            type_dict[output_def.runtime_type.name] = output_def.runtime_type
            for inner_type in output_def.runtime_type.inner_types:
                type_dict[inner_type.name] = inner_type

    return type_dict



from __future__ import absolute_import

import copy
import logging
import traceback

from collections import namedtuple
from logging import CRITICAL, DEBUG, ERROR, INFO, WARNING

import coloredlogs

from dagster import check, seven
from dagster.core.types.config import Enum, EnumValue

VALID_LEVELS = set([CRITICAL, DEBUG, ERROR, INFO, WARNING])

LOOKUP = {'CRITICAL': CRITICAL, 'DEBUG': DEBUG, 'ERROR': ERROR, 'INFO': INFO, 'WARNING': WARNING}

VALID_LEVEL_STRINGS = ['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL']

LogLevelEnum = Enum('log_level', list(map(EnumValue, VALID_LEVEL_STRINGS)))


def level_from_string(string):
    check.str_param(string, 'string')
    return LOOKUP[string]


class JsonFileHandler(logging.Handler):
    def __init__(self, json_path):
        super(JsonFileHandler, self).__init__()
        self.json_path = check.str_param(json_path, 'json_path')

    def emit(self, record):
        try:
            log_dict = copy.copy(record.__dict__)

            # This horrific monstrosity is to maintain backwards compatability
            # with the old behavior of the JsonFileHandler, which the clarify
            # project has a dependency on. It relied on the dagster-defined
            # properties smashing all the properties of the LogRecord object
            # and uploads all of those properties to a redshift table for
            # in order to do analytics on the log

            if 'dagster_meta' in log_dict:
                dagster_meta_dict = log_dict['dagster_meta']
                del log_dict['dagster_meta']
            else:
                dagster_meta_dict = {}

            log_dict.update(dagster_meta_dict)

            with open(self.json_path, 'a') as ff:
                text_line = seven.json.dumps(log_dict)
                ff.write(text_line + '\n')
        # Need to catch Exception here, so disabling lint
        except Exception as e:  # pylint: disable=W0703
            logging.critical('Error during logging!')
            logging.exception(str(e))


class StructuredLoggerMessage(
    namedtuple('_StructuredLoggerMessage', 'name message level meta record')
):
    def __new__(cls, name, message, level, meta, record):
        return super(StructuredLoggerMessage, cls).__new__(
            cls,
            check.str_param(name, 'name'),
            check.str_param(message, 'message'),
            check_valid_level_param(level),
            check.dict_param(meta, 'meta'),
            check.inst_param(record, 'record', logging.LogRecord),
        )


class JsonEventLoggerHandler(logging.Handler):
    def __init__(self, json_path, construct_event_record):
        super(JsonEventLoggerHandler, self).__init__()
        self.json_path = check.str_param(json_path, 'json_path')
        self.construct_event_record = construct_event_record

    def emit(self, record):
        try:
            event_record = self.construct_event_record(record)
            with open(self.json_path, 'a') as ff:
                text_line = seven.json.dumps(event_record.to_dict())
                ff.write(text_line + '\n')

        # Need to catch Exception here, so disabling lint
        except Exception as e:  # pylint: disable=W0703
            logging.critical('Error during logging!')
            logging.exception(str(e))


class StructuredLoggerHandler(logging.Handler):
    def __init__(self, callback):
        super(StructuredLoggerHandler, self).__init__()
        self.callback = check.is_callable(callback, 'callback')

    def emit(self, record):
        try:
            self.callback(
                StructuredLoggerMessage(
                    name=record.name,
                    message=record.msg,
                    level=record.levelno,
                    meta=record.dagster_meta,
                    record=record,
                )
            )
        # Need to catch Exception here, so disabling lint
        except Exception as e:  # pylint: disable=W0703
            logging.critical('Error during logging!')
            logging.exception(str(e))


def check_valid_level_param(level):
    check.param_invariant(
        level in VALID_LEVELS,
        'level',
        'Must be valid python logging level. Got {level}'.format(level=level),
    )
    return level


def construct_single_handler_logger(name, level, handler):
    check.str_param(name, 'name')
    check_valid_level_param(level)
    check.inst_param(handler, 'handler', logging.Handler)

    klass = logging.getLoggerClass()
    logger = klass(name, level=level)
    logger.addHandler(handler)
    return logger


def define_structured_logger(name, callback, level):
    check.str_param(name, 'name')
    check.callable_param(callback, 'callback')
    check_valid_level_param(level)

    return construct_single_handler_logger(name, level, StructuredLoggerHandler(callback))


def define_json_file_logger(name, json_path, level):
    check.str_param(name, 'name')
    check.str_param(json_path, 'json_path')
    check_valid_level_param(level)

    stream_handler = JsonFileHandler(json_path)
    stream_handler.setFormatter(define_default_formatter())
    return construct_single_handler_logger(name, level, stream_handler)


def define_colored_console_logger(name, level=INFO):
    check.str_param(name, 'name')
    check.param_invariant(
        level in VALID_LEVELS,
        'level',
        'Must be valid python logging level. Got {level}'.format(level=level),
    )

    klass = logging.getLoggerClass()
    logger = klass(name, level=level)
    coloredlogs.install(logger=logger, level=level, fmt=default_format_string())
    return logger


def default_format_string():
    return '%(asctime)s - %(name)s - %(levelname)s - %(message)s'


def define_default_formatter():
    return logging.Formatter(default_format_string())


def debug_format_string():
    return '''%(name)s.%(levelname)s: %(message)s
    time: %(asctime)s relative: %(relativeCreated)dms
    path: %(pathname)s line: %(lineno)d'''


def define_debug_formatter():
    return logging.Formatter(debug_format_string())


def get_formatted_stack_trace(exception):
    check.inst_param(exception, 'exception', Exception)
    return ''.join(get_stack_trace_array(exception))


def get_stack_trace_array(exception):
    check.inst_param(exception, 'exception', Exception)
    if hasattr(exception, '__traceback__'):
        tb = exception.__traceback__
    else:
        import sys

        _exc_type, _exc_value, tb = sys.exc_info()
    return traceback.format_tb(tb)



from collections import namedtuple
from enum import Enum

from dagster import check

from dagster.core.types import Field, String
from dagster.core.types.field_utils import check_user_facing_opt_field_param


class ResourceDefinition(object):
    '''Resources are pipeline-scoped ways to make external resources (like database connections)
    available to solids during pipeline execution and clean up after execution resolves.

    Args:
        resource_fn (Callable[[InitResourceContext], Any]):
            User provided function to instantiate the resource. This resource will be available to
            solids via ``context.resources``
        config_field (Field):
            The type for the configuration data for this resource, passed to ``resource_fn`` via
            ``init_context.resource_config``
        description (str)
    '''

    def __init__(self, resource_fn, config_field=None, description=None):
        self.resource_fn = check.callable_param(resource_fn, 'resource_fn')
        self.config_field = check_user_facing_opt_field_param(
            config_field, 'config_field', 'of a ResourceDefinition or @resource'
        )
        self.description = check.opt_str_param(description, 'description')

    @staticmethod
    def none_resource(description=None):
        return ResourceDefinition(resource_fn=lambda _init_context: None, description=description)

    @staticmethod
    def string_resource(description=None):
        return ResourceDefinition(
            resource_fn=lambda init_context: init_context.resource_config,
            config_field=Field(String),
            description=description,
        )


def resource(config_field=None, description=None):
    '''A decorator for creating a resource. The decorated function will be used as the
    resource_fn in a ResourceDefinition.
    '''

    # This case is for when decorator is used bare, without arguments.
    # E.g. @resource versus @resource()
    if callable(config_field):
        return ResourceDefinition(resource_fn=config_field)

    def _wrap(resource_fn):
        return ResourceDefinition(resource_fn, config_field, description)

    return _wrap


class ResourcesSource(Enum):
    '''Resources can either be created via a pipeline context definition:

    context=PipelineContextDefinition(resources={'foo': <some resource>})

    --or-- (less commonly) via a custom execution context:

    PipelineContextDefinition(
        context_fn=lambda init_context: ExecutionContext(
            resources=<something>
        ),
    )
    '''

    PIPELINE_CONTEXT_DEF = 'PIPELINE_CONTEXT_DEF'
    CUSTOM_EXECUTION_CONTEXT = 'CUSTOM_EXECUTION_CONTEXT'


class ResourcesBuilder(namedtuple('ResourcesBuilder', 'src resources_type')):
    def __new__(cls, src, resources_type):
        return super(ResourcesBuilder, cls).__new__(cls, src, resources_type)

    def build(self, mapper_fn=None, resource_deps=None):
        '''We dynamically create a type that has the resource keys as properties, to enable dotting into
        the resources from a context.

        For example, given:

        resources = {'foo': <some resource>, 'bar': <some other resource>}

        then this will create the type Resource(namedtuple('foo bar'))

        and then binds the specified resources into an instance of this object, which can be consumed
        as, e.g., context.resources.foo.
        '''
        if self.resources_type == ResourcesSource.PIPELINE_CONTEXT_DEF and self.src is not None:
            src = mapper_fn(self.src, resource_deps) if (mapper_fn and resource_deps) else self.src

            resource_type = namedtuple('Resources', list(src.keys()))
            return resource_type(**src)
        elif self.resources_type == ResourcesSource.CUSTOM_EXECUTION_CONTEXT:
            return self.src
        else:
            return None



import sys

from future.utils import raise_from

from dagster import check

from dagster.utils.timing import time_execution_scope

from dagster.core.errors import (
    DagsterError,
    DagsterInvariantViolationError,
    DagsterRuntimeCoercionError,
    DagsterExecutionStepExecutionError,
    DagsterTypeError,
    DagsterStepOutputNotFoundError,
    user_code_error_boundary,
)

from dagster.core.execution.execution_context import (
    ExecutorConfig,
    SystemPipelineExecutionContext,
    SystemStepExecutionContext,
)

from dagster.core.definitions import Materialization, ExpectationResult

from dagster.core.events import DagsterEvent, DagsterEventType

from dagster.core.intermediates_manager import IntermediatesManager

from dagster.utils.error import serializable_error_info_from_exc_info

from dagster.core.execution_plan.objects import (
    ExecutionStep,
    StepOutputHandle,
    StepOutputValue,
    StepOutputData,
    StepFailureData,
    StepSuccessData,
)

from dagster.core.execution_plan.plan import ExecutionPlan

from .engine_base import BaseEngine


class InProcessEngine(BaseEngine):  # pylint: disable=no-init
    @staticmethod
    def execute(pipeline_context, execution_plan, step_keys_to_execute=None):
        check.inst_param(pipeline_context, 'pipeline_context', SystemPipelineExecutionContext)
        check.inst_param(execution_plan, 'execution_plan', ExecutionPlan)
        check.opt_list_param(step_keys_to_execute, 'step_keys_to_execute', of_type=str)

        step_key_set = None if step_keys_to_execute is None else set(step_keys_to_execute)

        check.param_invariant(
            isinstance(pipeline_context.executor_config, ExecutorConfig),
            'pipeline_context',
            'Expected executor_config to be ExecutorConfig got {}'.format(
                pipeline_context.executor_config
            ),
        )

        failed_or_skipped_steps = set()

        step_levels = execution_plan.topological_step_levels()

        intermediates_manager = pipeline_context.intermediates_manager

        # It would be good to implement a reference tracking algorithm here so we could
        # garbage collection results that are no longer needed by any steps
        # https://github.com/dagster-io/dagster/issues/811

        for step_level in step_levels:
            for step in step_level:
                if step_key_set and step.key not in step_key_set:
                    continue

                step_context = pipeline_context.for_step(step)

                failed_inputs = [
                    step_input.prev_output_handle.step_key
                    for step_input in step.step_inputs
                    if step_input.prev_output_handle.step_key in failed_or_skipped_steps
                ]
                if failed_inputs:
                    step_context.log.info(
                        'Dependencies for step {step} failed: {failed_inputs}. Not executing.'.format(
                            step=step.key, failed_inputs=failed_inputs
                        )
                    )
                    failed_or_skipped_steps.add(step.key)
                    yield DagsterEvent.step_skipped_event(step_context)
                    continue

                uncovered_inputs = intermediates_manager.uncovered_inputs(step_context, step)
                if uncovered_inputs:
                    # In partial pipeline execution, we may end up here without having validated the
                    # missing dependent outputs were optional
                    _assert_missing_inputs_optional(uncovered_inputs, execution_plan, step.key)

                    step_context.log.info(
                        (
                            'Not all inputs covered for {step}. Not executing. Output missing for '
                            'inputs: {uncovered_inputs}'
                        ).format(uncovered_inputs=uncovered_inputs, step=step.key)
                    )
                    failed_or_skipped_steps.add(step.key)
                    yield DagsterEvent.step_skipped_event(step_context)
                    continue

                input_values = _create_input_values(step_context, intermediates_manager)

                for step_event in check.generator(
                    execute_step_in_memory(step_context, input_values, intermediates_manager)
                ):
                    check.inst(step_event, DagsterEvent)
                    if step_event.is_step_failure:
                        failed_or_skipped_steps.add(step.key)

                    yield step_event


def _assert_missing_inputs_optional(uncovered_inputs, execution_plan, step_key):
    nonoptionals = [
        handle for handle in uncovered_inputs if not execution_plan.get_step_output(handle).optional
    ]
    if nonoptionals:
        raise DagsterStepOutputNotFoundError(
            (
                'When executing {step} discovered required outputs missing '
                'from previous step: {nonoptionals}'
            ).format(nonoptionals=nonoptionals, step=step_key),
            step_key=nonoptionals[0].step_key,
            output_name=nonoptionals[0].output_name,
        )


def _create_input_values(step_context, intermediates_manager):
    check.inst_param(intermediates_manager, 'intermediates_manager', IntermediatesManager)

    step = step_context.step

    input_values = {}
    for step_input in step.step_inputs:
        if step_input.runtime_type.is_nothing:
            continue

        prev_output_handle = step_input.prev_output_handle
        input_value = intermediates_manager.get_intermediate(
            step_context, step_input.runtime_type, prev_output_handle
        )
        input_values[step_input.name] = input_value
    return input_values


def execute_step_in_memory(step_context, inputs, intermediates_manager):
    check.inst_param(step_context, 'step_context', SystemStepExecutionContext)
    check.dict_param(inputs, 'inputs', key_type=str)
    check.inst_param(intermediates_manager, 'intermediates_manager', IntermediatesManager)

    try:
        for step_event in check.generator(
            _execute_steps_core_loop(step_context, inputs, intermediates_manager)
        ):
            if step_event.event_type is DagsterEventType.STEP_OUTPUT:
                step_context.log.info(
                    'Step {step} emitted {value} for output {output}'.format(
                        step=step_context.step.key,
                        value=step_event.step_output_data.value_repr,
                        output=step_event.step_output_data.output_name,
                    )
                )
            yield step_event
    except DagsterError as dagster_error:
        user_facing_exc_info = (
            # pylint does not know original_exc_info exists is is_user_code_error is true
            # pylint: disable=no-member
            dagster_error.original_exc_info
            if dagster_error.is_user_code_error
            else sys.exc_info()
        )

        error_info = serializable_error_info_from_exc_info(user_facing_exc_info)

        yield DagsterEvent.step_failure_event(
            step_context=step_context, step_failure_data=StepFailureData(error=error_info)
        )

        if step_context.executor_config.raise_on_error:
            raise dagster_error

        return
    except:
        error_info = serializable_error_info_from_exc_info(sys.exc_info())
        yield DagsterEvent.step_failure_event(
            step_context=step_context, step_failure_data=StepFailureData(error=error_info)
        )
        raise


def _error_check_step_outputs(step, step_output_iter):
    check.inst_param(step, 'step', ExecutionStep)
    check.generator_param(step_output_iter, 'step_output_iter')

    output_names = list([output_def.name for output_def in step.step_outputs])
    seen_outputs = set()

    for step_output in step_output_iter:
        if not isinstance(step_output, StepOutputValue):
            yield step_output
            continue

        # do additional processing on StepOutputValues
        step_output_value = step_output
        if not step.has_step_output(step_output_value.output_name):
            raise DagsterInvariantViolationError(
                'Core transform for solid "{step.solid_handle.name}" returned an output '
                '"{step_output_value.output_name}" that does not exist. The available '
                'outputs are {output_names}'.format(
                    step=step, step_output_value=step_output_value, output_names=output_names
                )
            )

        if step_output_value.output_name in seen_outputs:
            raise DagsterInvariantViolationError(
                'Core transform for solid "{step.solid_handle.name}" returned an output '
                '"{step_output_value.output_name}" multiple times'.format(
                    step=step, step_output_value=step_output_value
                )
            )

        yield step_output_value
        seen_outputs.add(step_output_value.output_name)

    for step_output_def in step.step_outputs:
        if not step_output_def.name in seen_outputs and not step_output_def.optional:
            raise DagsterStepOutputNotFoundError(
                'Core transform for solid "{step.solid_handle.name}" did not return an output '
                'for non-optional output "{step_output_def.name}"'.format(
                    step=step, step_output_def=step_output_def
                ),
                step_key=step.key,
                output_name=step_output_def.name,
            )


def _execute_steps_core_loop(step_context, inputs, intermediates_manager):
    check.inst_param(step_context, 'step_context', SystemStepExecutionContext)
    check.dict_param(inputs, 'inputs', key_type=str)
    check.inst_param(intermediates_manager, 'intermediates_manager', IntermediatesManager)

    evaluated_inputs = {}
    # do runtime type checks of inputs versus step inputs
    for input_name, input_value in inputs.items():
        evaluated_inputs[input_name] = _get_evaluated_input(
            step_context.step, input_name, input_value
        )

    yield DagsterEvent.step_start_event(step_context)

    with time_execution_scope() as timer_result:
        step_output_iterator = check.generator(
            _iterate_step_outputs_within_boundary(step_context, evaluated_inputs)
        )
    for step_output in check.generator(
        _error_check_step_outputs(step_context.step, step_output_iterator)
    ):

        if isinstance(step_output, StepOutputValue):
            yield _create_step_output_event(step_context, step_output, intermediates_manager)
        elif isinstance(step_output, Materialization):
            yield DagsterEvent.step_materialization(step_context, step_output)
        elif isinstance(step_output, ExpectationResult):
            yield DagsterEvent.step_expectation_result(step_context, step_output)
        else:
            check.failed(
                'Unexpected step_output {step_output}, should have been caught earlier'.format(
                    step_output=step_output
                )
            )

    yield DagsterEvent.step_success_event(
        step_context, StepSuccessData(duration_ms=timer_result.millis)
    )


def _create_step_output_event(step_context, step_output_value, intermediates_manager):
    check.inst_param(step_context, 'step_context', SystemStepExecutionContext)
    check.inst_param(step_output_value, 'step_output_value', StepOutputValue)
    check.inst_param(intermediates_manager, 'intermediates_manager', IntermediatesManager)

    step = step_context.step
    step_output = step.step_output_named(step_output_value.output_name)

    try:
        value = step_output.runtime_type.coerce_runtime_value(step_output_value.value)
        step_output_handle = StepOutputHandle.from_step(
            step=step, output_name=step_output_value.output_name
        )

        object_key = intermediates_manager.set_intermediate(
            context=step_context,
            runtime_type=step_output.runtime_type,
            step_output_handle=step_output_handle,
            value=value,
        )

        return DagsterEvent.step_output_event(
            step_context=step_context,
            step_output_data=StepOutputData(
                step_output_handle=step_output_handle,
                value_repr=repr(value),
                intermediate_materialization=Materialization(path=object_key)
                if object_key
                else None,
            ),
        )
    except DagsterRuntimeCoercionError as e:
        raise DagsterInvariantViolationError(
            (
                'In solid "{step.solid_handle.name}" the output "{output_name}" returned '
                'an invalid type: {error_msg}.'
            ).format(
                step=step, error_msg=','.join(e.args), output_name=step_output_value.output_name
            )
        )


def _get_evaluated_input(step, input_name, input_value):
    check.inst_param(step, 'step', ExecutionStep)
    check.str_param(input_name, 'input_name')

    step_input = step.step_input_named(input_name)
    try:
        return step_input.runtime_type.coerce_runtime_value(input_value)
    except DagsterRuntimeCoercionError as evaluate_error:
        raise_from(
            DagsterTypeError(
                (
                    'In solid "{step.solid_handle.name}" the input "{input_name}" received value {input_value} '
                    'of Python type {input_type} which does not pass the typecheck for '
                    'Dagster type {step_input.runtime_type.name}. Step {step.key}.'
                ).format(
                    step=step,
                    input_name=input_name,
                    input_value=input_value,
                    input_type=type(input_value),
                    step_input=step_input,
                )
            ),
            evaluate_error,
        )


def _iterate_step_outputs_within_boundary(step_context, evaluated_inputs):
    check.inst_param(step_context, 'step_context', SystemStepExecutionContext)
    check.dict_param(evaluated_inputs, 'evaluated_inputs', key_type=str)

    error_str = '''Error occured during the execution of step:
    step key: "{key}"
    solid instance: "{solid}"
    solid definition: "{solid_def}"
    '''.format(
        key=step_context.step.key,
        solid_def=step_context.solid_def.name,
        solid=step_context.solid.name,
    )

    with user_code_error_boundary(
        DagsterExecutionStepExecutionError,
        error_str,
        step_key=step_context.step.key,
        solid_def_name=step_context.solid_def.name,
        solid_name=step_context.solid.name,
    ):
        gen = check.opt_generator(step_context.step.compute_fn(step_context, evaluated_inputs))

        if gen is not None:
            for step_output in gen:
                yield step_output



'''
.. data:: ANY
    TEst test test
'''

from enum import Enum


class BuiltinEnum(Enum):

    ANY = 'Any'
    BOOL = 'Bool'
    FLOAT = 'Float'
    INT = 'Int'
    PATH = 'Path'
    STRING = 'String'
    NOTHING = 'Nothing'



from dagster import check

from dagster.core.definitions import ModeDefinition, PipelineContextDefinition
from dagster.core.system_config.objects import EnvironmentConfig

from .user_context import ExecutionContext


class ResourceCreationAdapter:
    '''
    This class exists temporarily to handle the transition period where we have three ways
    of creating resources:
    1) Override in a context function
    2) Specifiying a set of resources on a context definition
    3) Specifiying a set of resources on a mode definition

    When we are done with this refactor only 3 will be necessary. The elimination of this
    class will be a good marker of when this is "done" -- schrockn (05-08-2019)
    '''

    def __init__(self, execution_context, environment_config, context_definition, mode_definition):
        self.execution_context = check.inst_param(
            execution_context, 'execution_context', ExecutionContext
        )

        self.environment_config = check.inst_param(
            environment_config, 'environment_config', EnvironmentConfig
        )

        check.invariant(context_definition or mode_definition)
        check.invariant(not (context_definition and mode_definition))

        self.context_definition = check.opt_inst_param(
            context_definition, 'context_definition', PipelineContextDefinition
        )

        self.mode_definition = check.opt_inst_param(
            mode_definition, 'mode_definition', ModeDefinition
        )

    @property
    def is_resource_override(self):
        return bool(self.execution_context.resources)

    def get_override_resources(self):
        check.invariant(self.is_resource_override)
        return self.execution_context.resources

    def get_resource_config(self, resource_name):
        if self.context_definition:
            check.invariant(self.environment_config.resources is None)
            return self.environment_config.context.resources.get(resource_name, {}).get('config')
        else:
            check.invariant(self.environment_config.context is None)
            return self.environment_config.resources.get(resource_name, {}).get('config')

    @property
    def resource_defs(self):
        return (
            self.context_definition.resource_defs
            if self.context_definition
            else self.mode_definition.resource_defs
        )



from dagster.tutorials.intro_tutorial.unittesting import (
    execute_test_only_final,
    execute_test_a_plus_b_final_subdag,
)


def test_only_final():
    execute_test_only_final()


def test_a_plus_b_final_subdag():
    execute_test_a_plus_b_final_subdag()



import os
import shutil

from abc import ABCMeta, abstractmethod

import six

from dagster import check, seven
from dagster.utils import mkdir_p

from dagster.core.execution.execution_context import SystemPipelineExecutionContext
from .runs import RunStorageMode
from .types.runtime import RuntimeType, resolve_to_runtime_type


class TypeStoragePlugin(six.with_metaclass(ABCMeta)):  # pylint: disable=no-init
    '''Base class for storage plugins.

    Extend this class for (storage_mode, runtime_type) pairs that need special handling.
    '''

    @classmethod
    @abstractmethod
    def set_object(cls, object_store, obj, context, runtime_type, paths):
        check.subclass_param(object_store, 'object_store', ObjectStore)
        return object_store.set_object(obj, context, runtime_type, paths)

    @classmethod
    @abstractmethod
    def get_object(cls, object_store, context, runtime_type, paths):
        check.subclass_param(object_store, 'object_store', ObjectStore)
        return object_store.get_object(context, runtime_type, paths)


class ObjectStore(six.with_metaclass(ABCMeta)):
    def __init__(self, types_to_register=None):
        types_to_register = check.opt_dict_param(
            types_to_register,
            'types_to_register',
            key_type=RuntimeType,
            value_class=TypeStoragePlugin,
        )
        self.TYPE_STORAGE_PLUGIN_REGISTRY = {}

        for type_to_register, type_storage_plugin in types_to_register.items():
            self.register_type(type_to_register, type_storage_plugin)

    def register_type(self, type_to_register, type_storage_plugin):
        check.inst_param(type_to_register, 'type_to_register', RuntimeType)
        check.subclass_param(type_storage_plugin, 'type_storage_plugin', TypeStoragePlugin)
        check.invariant(
            type_to_register.name is not None,
            'Cannot register a type storage plugin for an anonymous type',
        )
        self.TYPE_STORAGE_PLUGIN_REGISTRY[type_to_register.name] = type_storage_plugin

    @abstractmethod
    def set_object(self, obj, context, runtime_type, paths):
        pass

    @abstractmethod
    def get_object(self, context, runtime_type, paths):
        pass

    @abstractmethod
    def has_object(self, context, paths):
        pass

    @abstractmethod
    def rm_object(self, context, paths):
        pass

    def _check_for_unsupported_composite_overrides(self, runtime_type):
        composite_overrides = {
            t.name for t in runtime_type.inner_types if t.name in self.TYPE_STORAGE_PLUGIN_REGISTRY
        }
        if composite_overrides:
            outer_type = 'composite type'
            if runtime_type.is_list:
                if runtime_type.is_nullable:
                    outer_type = 'Nullable List'
                else:
                    outer_type = 'List'
            elif runtime_type.is_nullable:
                outer_type = 'Nullable'

            if len(composite_overrides) > 1:
                plural = 's'
                this = 'These'
                has = 'have'
            else:
                plural = ''
                this = 'This'
                has = 'has'

            check.not_implemented(
                'You are attempting to store a {outer_type} containing type{plural} '
                '{type_names} in a object store. {this} type{plural} {has} specialized storage '
                'behavior (configured in the TYPE_STORAGE_PLUGIN_REGISTRY). We do not '
                'currently support storing Nullables or Lists of types with customized '
                'storage. See https://github.com/dagster-io/dagster/issues/1190 for '
                'details.'.format(
                    outer_type=outer_type,
                    plural=plural,
                    this=this,
                    has=has,
                    type_names=', '.join([str(x) for x in composite_overrides]),
                )
            )

    def set_value(self, obj, context, runtime_type, paths):
        if runtime_type.name is not None and runtime_type.name in self.TYPE_STORAGE_PLUGIN_REGISTRY:
            return self.TYPE_STORAGE_PLUGIN_REGISTRY[runtime_type.name].set_object(
                self, obj, context, runtime_type, paths
            )
        elif runtime_type.name is None:
            self._check_for_unsupported_composite_overrides(runtime_type)
        return self.set_object(obj, context, runtime_type, paths)

    def get_value(self, context, runtime_type, paths):
        if runtime_type.name is not None and runtime_type.name in self.TYPE_STORAGE_PLUGIN_REGISTRY:
            return self.TYPE_STORAGE_PLUGIN_REGISTRY[runtime_type.name].get_object(
                self, context, runtime_type, paths
            )
        elif runtime_type.name is None:
            self._check_for_unsupported_composite_overrides(runtime_type)
        return self.get_object(context, runtime_type, paths)


def get_run_files_directory(base_dir, run_id):
    return os.path.join(base_dir, 'dagster', 'runs', run_id, 'files')


def get_valid_target_path(base_dir, paths):
    if len(paths) > 1:
        target_dir = os.path.join(base_dir, *paths[:-1])
        mkdir_p(target_dir)
        return os.path.join(target_dir, paths[-1])
    else:
        check.invariant(len(paths) == 1)
        target_dir = base_dir
        mkdir_p(target_dir)
        return os.path.join(target_dir, paths[0])


class FileSystemObjectStore(ObjectStore):
    def __init__(self, run_id, types_to_register=None, base_dir=None):
        self.run_id = check.str_param(run_id, 'run_id')
        self.storage_mode = RunStorageMode.FILESYSTEM
        self._base_dir = os.path.abspath(
            os.path.expanduser(
                check.opt_nonempty_str_param(
                    base_dir, 'base_dir', seven.get_system_temp_directory()
                )
            )
        )
        check.invariant(
            os.path.isdir(self._base_dir),
            'Could not find a directory at the base_dir supplied to FileSystemObjectStore: '
            '{base_dir}'.format(base_dir=self._base_dir),
        )
        self.root = get_run_files_directory(self.base_dir, run_id)

        super(FileSystemObjectStore, self).__init__(types_to_register)

    @property
    def base_dir(self):
        return self._base_dir

    def url_for_paths(self, paths):
        return 'file:///' + '/'.join([self.root] + paths)

    def set_object(self, obj, context, runtime_type, paths):  # pylint: disable=unused-argument
        check.inst_param(context, 'context', SystemPipelineExecutionContext)
        check.inst_param(runtime_type, 'runtime_type', RuntimeType)
        check.list_param(paths, 'paths', of_type=str)
        check.param_invariant(len(paths) > 0, 'paths')

        target_path = get_valid_target_path(self.root, paths)

        check.invariant(not os.path.exists(target_path))
        # This is not going to be right in the general case, e.g. for types like Spark
        # datasets/dataframes, which naturally serialize to
        # union(parquet_file, directory([parquet_file])) -- we will need a) to pass the
        # object store into the serializer and b) to provide sugar for the common case where
        # we don't need to do anything other than open the target path as a binary file
        with open(target_path, 'wb') as ff:
            runtime_type.serialization_strategy.serialize_value(context, obj, ff)

        return target_path

    def get_object(self, context, runtime_type, paths):  # pylint: disable=unused-argument
        check.list_param(paths, 'paths', of_type=str)
        check.inst_param(runtime_type, 'runtime_type', RuntimeType)

        check.param_invariant(len(paths) > 0, 'paths')
        target_path = os.path.join(self.root, *paths)
        with open(target_path, 'rb') as ff:
            return runtime_type.serialization_strategy.deserialize_value(context, ff)

    def has_object(self, context, paths):  # pylint: disable=unused-argument
        target_path = os.path.join(self.root, *paths)
        return os.path.exists(target_path)

    def rm_object(self, context, paths):  # pylint: disable=unused-argument
        target_path = os.path.join(self.root, *paths)
        if not self.has_object(context, paths):
            return
        if os.path.isfile(target_path):
            os.unlink(target_path)
        elif os.path.isdir(target_path):
            shutil.rmtree(target_path)
        return

    def copy_object_from_prev_run(
        self, context, previous_run_id, paths
    ):  # pylint: disable=unused-argument
        prev_run_files_dir = get_run_files_directory(self.base_dir, previous_run_id)
        check.invariant(os.path.isdir(prev_run_files_dir))

        copy_from_path = os.path.join(prev_run_files_dir, *paths)
        copy_to_path = get_valid_target_path(self.root, paths)

        check.invariant(
            not os.path.exists(copy_to_path), 'Path already exists {}'.format(copy_to_path)
        )

        if os.path.isfile(copy_from_path):
            shutil.copy(copy_from_path, copy_to_path)
        elif os.path.isdir(copy_from_path):
            shutil.copytree(copy_from_path, copy_to_path)
        else:
            check.failed('should not get here')


def get_fs_paths(step_key, output_name):
    return ['intermediates', step_key, output_name]


def get_filesystem_intermediate(run_id, step_key, dagster_type, output_name='result'):
    object_store = FileSystemObjectStore(run_id)
    return object_store.get_object(
        context=None,
        runtime_type=resolve_to_runtime_type(dagster_type),
        paths=get_fs_paths(step_key, output_name),
    )


def has_filesystem_intermediate(run_id, step_key, output_name='result'):
    object_store = FileSystemObjectStore(run_id)
    return object_store.has_object(context=None, paths=get_fs_paths(step_key, output_name))


def construct_type_storage_plugin_registry(pipeline_def, storage_mode):
    return {
        type_obj: type_obj.storage_plugins.get(storage_mode)
        for type_obj in pipeline_def.all_runtime_types()
        if type_obj.storage_plugins.get(storage_mode)
    }



# encoding: utf-8
# py27 compat

from dagster import Field, PipelineDefinition, execute_pipeline, solid, types


@solid(config_field=Field(types.String, is_optional=True, default_value='en-us'))
def configurable_hello(context):
    if len(context.solid_config) >= 3 and context.solid_config[:3] == 'haw':
        return 'Aloha honua!'
    elif len(context.solid_config) >= 2 and context.solid_config[:2] == 'cn':
        return '你好, 世界!'
    else:
        return 'Hello, world!'


def define_configurable_hello_pipeline():
    return PipelineDefinition(name='configurable_hello_pipeline', solids=[configurable_hello])


def test_intro_tutorial_part_four():
    execute_pipeline(
        define_configurable_hello_pipeline(), {'solids': {'configurable_hello': {'config': 'cn'}}}
    )



from dagster import (
    DependencyDefinition,
    Field,
    InputDefinition,
    MultipleResults,
    OutputDefinition,
    PipelineDefinition,
    Result,
    solid,
    String,
    Int,
)


@solid(
    outputs=[
        OutputDefinition(dagster_type=Int, name='out_one'),
        OutputDefinition(dagster_type=Int, name='out_two'),
    ]
)
def yield_outputs(_context):
    yield Result(23, 'out_one')
    yield Result(45, 'out_two')


@solid(
    outputs=[
        OutputDefinition(dagster_type=Int, name='out_one'),
        OutputDefinition(dagster_type=Int, name='out_two'),
    ]
)
def return_dict_results(_context):
    return MultipleResults.from_dict({'out_one': 23, 'out_two': 45})


@solid(
    config_field=Field(String, description='Should be either out_one or out_two'),
    outputs=[
        OutputDefinition(dagster_type=Int, name='out_one', is_optional=True),
        OutputDefinition(dagster_type=Int, name='out_two', is_optional=True),
    ],
)
def conditional(context):
    if context.solid_config == 'out_one':
        yield Result(23, 'out_one')
    elif context.solid_config == 'out_two':
        yield Result(45, 'out_two')
    else:
        raise Exception('invalid config')


@solid(inputs=[InputDefinition('num', dagster_type=Int)])
def log_num(context, num):
    context.log.info('num {num}'.format(num=num))
    return num


@solid(inputs=[InputDefinition('num', dagster_type=Int)])
def log_num_squared(context, num):
    context.log.info('num_squared {num_squared}'.format(num_squared=num * num))
    return num * num


def define_multiple_outputs_step_one_pipeline():
    return PipelineDefinition(
        name='multiple_outputs_step_one_pipeline',
        solids=[return_dict_results, log_num, log_num_squared],
        dependencies={
            'log_num': {'num': DependencyDefinition(solid='return_dict_results', output='out_one')},
            'log_num_squared': {
                'num': DependencyDefinition(solid='return_dict_results', output='out_two')
            },
        },
    )


def define_multiple_outputs_step_two_pipeline():
    return PipelineDefinition(
        name='multiple_outputs_step_two_pipeline',
        solids=[yield_outputs, log_num, log_num_squared],
        dependencies={
            'log_num': {'num': DependencyDefinition('yield_outputs', 'out_one')},
            'log_num_squared': {'num': DependencyDefinition('yield_outputs', 'out_two')},
        },
    )


def define_multiple_outputs_step_three_pipeline():
    return PipelineDefinition(
        name='multiple_outputs_step_three_pipeline',
        solids=[conditional, log_num, log_num_squared],
        dependencies={
            'log_num': {'num': DependencyDefinition('conditional', 'out_one')},
            'log_num_squared': {'num': DependencyDefinition('conditional', 'out_two')},
        },
    )



from collections import namedtuple

from dagster import check

from dagster.core.definitions import (
    solids_in_topological_order,
    InputDefinition,
    OutputDefinition,
    PipelineDefinition,
    Solid,
    SolidOutputHandle,
)

from dagster.core.errors import DagsterInvariantViolationError

from dagster.core.system_config.objects import EnvironmentConfig

from dagster.core.utils import toposort

from .expectations import create_expectations_subplan, decorate_with_expectations

from .input_thunk import create_input_thunk_execution_step

from .materialization_thunk import decorate_with_output_materializations

from .objects import ExecutionStep, ExecutionValueSubplan, StepInput, StepOutputHandle, StepKind

from .transform import create_transform_step


class _PlanBuilder:
    '''_PlanBuilder. This is the state that is built up during the execution plan build process.

    steps List[ExecutionStep]: a list of the execution steps that have been created.

    step_output_map Dict[SolidOutputHandle, StepOutputHandle]:  maps logical solid outputs
    (solid_name, output_name) to particular step outputs. This covers the case where a solid maps to
    multiple steps and one wants to be able to attach to the logical output of a solid during
    execution.
    '''

    def __init__(self):
        self.steps = []
        self.step_output_map = dict()
        self.seen_keys = set()

    def add_step(self, step):
        # Keep track of the step keys we've seen so far to ensure we don't add duplicates
        if step.key in self.seen_keys:
            keys = [s.key for s in self.steps]
            check.failed(
                'Duplicated key {key}. Full list seen so far: {key_list}.'.format(
                    key=step.key, key_list=keys
                )
            )
        self.seen_keys.add(step.key)
        self.steps.append(step)

    def add_steps(self, steps):
        for step in steps:
            self.add_step(step)

    def get_output_handle(self, key):
        check.inst_param(key, 'key', SolidOutputHandle)
        return self.step_output_map[key]

    def set_output_handle(self, key, val):
        check.inst_param(key, 'key', SolidOutputHandle)
        check.inst_param(val, 'val', StepOutputHandle)
        self.step_output_map[key] = val

    def build(self, pipeline_def, artifacts_persisted):
        '''Builds the execution plan.
        '''

        # Construct dependency dictionary
        deps = {step.key: set() for step in self.steps}

        for step in self.steps:
            for step_input in step.step_inputs:
                deps[step.key].add(step_input.prev_output_handle.step_key)

        step_dict = {step.key: step for step in self.steps}

        return ExecutionPlan(pipeline_def, step_dict, deps, artifacts_persisted)


def create_subplan_for_input(
    pipeline_def, environment_config, solid, prev_step_output_handle, input_def
):
    check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition)
    check.inst_param(environment_config, 'environment_config', EnvironmentConfig)
    check.inst_param(solid, 'solid', Solid)
    check.inst_param(prev_step_output_handle, 'prev_step_output_handle', StepOutputHandle)
    check.inst_param(input_def, 'input_def', InputDefinition)

    if environment_config.expectations.evaluate and input_def.expectations:
        return create_expectations_subplan(
            pipeline_def, solid, input_def, prev_step_output_handle, kind=StepKind.INPUT_EXPECTATION
        )
    else:
        return ExecutionValueSubplan.empty(prev_step_output_handle)


def create_subplan_for_output(
    pipeline_def, environment_config, solid, solid_transform_step, output_def
):
    check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition)
    check.inst_param(solid, 'solid', Solid)
    check.inst_param(solid_transform_step, 'solid_transform_step', ExecutionStep)
    check.inst_param(output_def, 'output_def', OutputDefinition)

    subplan = decorate_with_expectations(
        pipeline_def, environment_config, solid, solid_transform_step, output_def
    )

    return decorate_with_output_materializations(
        pipeline_def, environment_config, solid, output_def, subplan
    )


def get_input_source_step_handle(
    pipeline_def, environment_config, plan_builder, solid, input_name, input_def
):
    check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition)
    check.inst_param(environment_config, 'environment_config', EnvironmentConfig)
    check.inst_param(plan_builder, 'plan_builder', _PlanBuilder)
    check.inst_param(solid, 'solid', Solid)
    check.str_param(input_name, 'input_name')
    check.inst_param(input_def, 'input_def', InputDefinition)

    input_handle = solid.input_handle(input_name)
    solid_config = environment_config.solids.get(solid.name)
    dependency_structure = pipeline_def.dependency_structure
    if solid_config and input_def.name in solid_config.inputs:
        step_creation_data = create_input_thunk_execution_step(
            pipeline_def, solid, input_def, solid_config.inputs[input_name]
        )
        plan_builder.add_step(step_creation_data.step)
        return step_creation_data.step_output_handle
    elif dependency_structure.has_dep(input_handle):
        solid_output_handle = dependency_structure.get_dep(input_handle)
        return plan_builder.get_output_handle(solid_output_handle)
    elif input_def.runtime_type.is_nothing:
        return None
    else:
        raise DagsterInvariantViolationError(
            (
                'In pipeline {pipeline_name} solid {solid_name}, input {input_name} '
                'must get a value either (a) from a dependency or (b) from the '
                'inputs section of its configuration.'
            ).format(pipeline_name=pipeline_def.name, solid_name=solid.name, input_name=input_name)
        )


class ExecutionPlan(
    namedtuple('_ExecutionPlan', 'pipeline_def step_dict deps steps artifacts_persisted')
):
    def __new__(cls, pipeline_def, step_dict, deps, artifacts_persisted):
        return super(ExecutionPlan, cls).__new__(
            cls,
            pipeline_def=check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition),
            step_dict=check.dict_param(
                step_dict, 'step_dict', key_type=str, value_type=ExecutionStep
            ),
            deps=check.dict_param(deps, 'deps', key_type=str, value_type=set),
            steps=list(step_dict.values()),
            artifacts_persisted=check.bool_param(artifacts_persisted, 'artifacts_persisted'),
        )

    def get_step_output(self, step_output_handle):
        check.inst_param(step_output_handle, 'step_output_handle', StepOutputHandle)
        step = self.get_step_by_key(step_output_handle.step_key)
        return step.step_output_named(step_output_handle.output_name)

    def has_step(self, key):
        check.str_param(key, 'key')
        return key in self.step_dict

    def get_step_by_key(self, key):
        check.str_param(key, 'key')
        return self.step_dict[key]

    def topological_steps(self):
        return [step for step_level in self.topological_step_levels() for step in step_level]

    def topological_step_levels(self):
        return [
            [self.step_dict[step_key] for step_key in step_key_level]
            for step_key_level in toposort(self.deps)
        ]

    @staticmethod
    def build(pipeline_def, environment_config):
        '''Here we build a new ExecutionPlan from a pipeline definition and the environment config.

        To do this, we iterate through the pipeline's solids in topological order, and hand off the
        execution steps for each solid to a companion _PlanBuilder object.

        Once we've processed the entire pipeline, we invoke _PlanBuilder.build() to construct the
        ExecutionPlan object.
        '''
        check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition)
        check.inst_param(environment_config, 'environment_config', EnvironmentConfig)

        plan_builder = _PlanBuilder()

        for solid in solids_in_topological_order(pipeline_def):
            ### 1. INPUTS
            # Create and add execution plan steps for solid inputs
            step_inputs = []
            for input_name, input_def in solid.definition.input_dict.items():
                prev_step_output_handle = get_input_source_step_handle(
                    pipeline_def, environment_config, plan_builder, solid, input_name, input_def
                )

                # We return None for the handle (see above in get_input_source_step_handle) when the
                # input def runtime type is "Nothing"
                if not prev_step_output_handle:
                    continue

                subplan = create_subplan_for_input(
                    pipeline_def, environment_config, solid, prev_step_output_handle, input_def
                )

                plan_builder.add_steps(subplan.steps)

                step_inputs.append(
                    StepInput(
                        input_def.name, input_def.runtime_type, subplan.terminal_step_output_handle
                    )
                )

            ### 2. TRANSFORM FUNCTION
            # Create and add execution plan step for the solid transform function
            solid_transform_step = create_transform_step(
                pipeline_def, environment_config, solid, step_inputs
            )
            plan_builder.add_step(solid_transform_step)

            ### 3. OUTPUTS
            # Create and add execution plan steps (and output handles) for solid outputs
            for name, output_def in solid.definition.output_dict.items():
                subplan = create_subplan_for_output(
                    pipeline_def, environment_config, solid, solid_transform_step, output_def
                )
                plan_builder.add_steps(subplan.steps)

                output_handle = solid.output_handle(name)
                plan_builder.set_output_handle(output_handle, subplan.terminal_step_output_handle)

        # Finally, we build and return the execution plan
        return plan_builder.build(
            pipeline_def=pipeline_def,
            artifacts_persisted=environment_config.storage.construct_run_storage().is_persistent,
        )



import pytest

from dagster import (
    PipelineDefinition,
    ModeDefinition,
    execute_pipeline,
    RunConfig,
    Field,
    Int,
    solid,
    resource,
    DagsterInvariantViolationError,
)


def test_basic_mode_definition():
    pipeline_def = PipelineDefinition(
        name='takes a mode', solids=[], mode_definitions=[ModeDefinition()]
    )
    assert pipeline_def


def test_mode_takes_a_name():
    pipeline_def = PipelineDefinition(
        name='takes a mode', solids=[], mode_definitions=[ModeDefinition(name='a_mode')]
    )
    assert pipeline_def


def define_modeless_pipeline():
    @solid
    def return_one(_context):
        return 1

    return PipelineDefinition(name='modeless', solids=[return_one])


def define_single_mode_pipeline():
    @solid
    def return_two(_context):
        return 2

    return PipelineDefinition(
        name='single_mode', solids=[return_two], mode_definitions=[ModeDefinition(name='the_mode')]
    )


def define_multi_mode_pipeline():
    @solid
    def return_three(_context):
        return 3

    return PipelineDefinition(
        name='multi_mode',
        solids=[return_three],
        mode_definitions=[ModeDefinition(name='mode_one'), ModeDefinition('mode_two')],
    )


def test_execute_modeless():
    pipeline_result = execute_pipeline(define_modeless_pipeline())
    assert pipeline_result.result_for_solid('return_one').transformed_value() == 1


def test_execute_single_mode():
    single_mode_pipeline = define_single_mode_pipeline()
    assert single_mode_pipeline.is_modeless is False
    assert single_mode_pipeline.is_single_mode is True

    assert (
        execute_pipeline(single_mode_pipeline).result_for_solid('return_two').transformed_value()
        == 2
    )

    assert (
        execute_pipeline(single_mode_pipeline, run_config=RunConfig(mode='the_mode'))
        .result_for_solid('return_two')
        .transformed_value()
        == 2
    )


def test_wrong_single_mode():
    with pytest.raises(DagsterInvariantViolationError):
        assert (
            execute_pipeline(define_single_mode_pipeline(), run_config=RunConfig(mode='wrong_mode'))
            .result_for_solid('return_two')
            .transformed_value()
            == 2
        )


def test_execute_multi_mode():
    multi_mode_pipeline = define_multi_mode_pipeline()

    assert (
        execute_pipeline(multi_mode_pipeline, run_config=RunConfig(mode='mode_one'))
        .result_for_solid('return_three')
        .transformed_value()
        == 3
    )

    assert (
        execute_pipeline(multi_mode_pipeline, run_config=RunConfig(mode='mode_two'))
        .result_for_solid('return_three')
        .transformed_value()
        == 3
    )


def test_execute_multi_mode_errors():
    multi_mode_pipeline = define_multi_mode_pipeline()

    with pytest.raises(DagsterInvariantViolationError):
        execute_pipeline(multi_mode_pipeline)

    with pytest.raises(DagsterInvariantViolationError):
        execute_pipeline(multi_mode_pipeline, run_config=RunConfig(mode='wrong_mode'))


def define_multi_mode_with_resources_pipeline():
    @resource(config_field=Field(Int))
    def adder_resource(init_context):
        return lambda x: x + init_context.resource_config

    @resource(config_field=Field(Int))
    def multer_resource(init_context):
        return lambda x: x * init_context.resource_config

    @solid
    def apply_to_three(context):
        return context.resources.op(3)

    return PipelineDefinition(
        name='multi_mode',
        solids=[apply_to_three],
        mode_definitions=[
            ModeDefinition(name='add_mode', resources={'op': adder_resource}),
            ModeDefinition(name='mult_mode', resources={'op': multer_resource}),
        ],
    )


def test_execute_multi_mode_with_resources():
    pipeline_def = define_multi_mode_with_resources_pipeline()

    add_mode_result = execute_pipeline(
        pipeline_def,
        run_config=RunConfig(mode='add_mode'),
        environment_dict={'resources': {'op': {'config': 2}}},
    )

    assert add_mode_result.result_for_solid('apply_to_three').transformed_value() == 5

    mult_mode_result = execute_pipeline(
        pipeline_def,
        run_config=RunConfig(mode='mult_mode'),
        environment_dict={'resources': {'op': {'config': 3}}},
    )

    assert mult_mode_result.result_for_solid('apply_to_three').transformed_value() == 9



from collections import namedtuple

from dagster import check


class Materialization(namedtuple('_Materialization', 'path description')):
    def __new__(cls, path, description=None):
        return super(Materialization, cls).__new__(
            cls,
            path=check.opt_str_param(path, 'path'),
            description=check.opt_str_param(description, 'description'),
        )



from dagster import PipelineDefinition, execute_pipeline, solid


@solid
def debug_message(context):
    context.log.debug('A debug message.')
    return 'foo'


@solid
def error_message(context):
    context.log.error('An error occurred.')


def define_execution_context_pipeline_step_one():
    return PipelineDefinition(solids=[debug_message, error_message])


def define_execution_context_pipeline_step_two():
    return PipelineDefinition(
        name='execution_context_pipeline', solids=[debug_message, error_message]
    )


def define_execution_context_pipeline_step_three():
    return PipelineDefinition(
        name='execution_context_pipeline', solids=[debug_message, error_message]
    )


if __name__ == '__main__':
    execute_pipeline(
        define_execution_context_pipeline_step_three(),
        {'context': {'default': {'config': {'log_level': 'DEBUG'}}}},
    )



from abc import ABCMeta, abstractmethod
import pickle

import six

from dagster import check


class SerializationStrategy(six.with_metaclass(ABCMeta)):  # pylint: disable=no-init
    @abstractmethod
    def serialize_value(self, context, value, write_file_obj):
        pass

    @abstractmethod
    def deserialize_value(self, context, read_file_obj):
        pass


class PickleSerializationStrategy(SerializationStrategy):  # pylint: disable=no-init
    def serialize_value(self, _context, value, write_file_obj):
        pickle.dump(value, write_file_obj)

    def deserialize_value(self, _context, read_file_obj):
        return pickle.load(read_file_obj)


def serialize_to_file(context, serialization_strategy, value, write_path):
    check.inst_param(serialization_strategy, 'serialization_strategy', SerializationStrategy)
    check.str_param(write_path, 'write_path')

    with open(write_path, 'wb') as write_obj:
        return serialization_strategy.serialize_value(context, value, write_obj)


def deserialize_from_file(context, serialization_strategy, read_path):
    check.inst_param(serialization_strategy, 'serialization_strategy', SerializationStrategy)
    check.str_param(read_path, 'read_path')

    with open(read_path, 'rb') as read_obj:
        return serialization_strategy.deserialize_value(context, read_obj)



from dagster import DependencyDefinition, InputDefinition, PipelineDefinition, lambda_solid


@lambda_solid
def solid_a():
    return 1


@lambda_solid(inputs=[InputDefinition('arg_a')])
def solid_b(arg_a):
    return arg_a * 2


@lambda_solid(inputs=[InputDefinition('arg_a')])
def solid_c(arg_a):
    return arg_a * 3


@lambda_solid(inputs=[InputDefinition('arg_b'), InputDefinition('arg_c')])
def solid_d(arg_b, arg_c):
    return arg_b * arg_c


def define_diamond_dag_pipeline():
    return PipelineDefinition(
        name='actual_dag_pipeline',
        # The order of this list does not matter:
        # dependencies determine execution order
        solids=[solid_d, solid_c, solid_b, solid_a],
        dependencies={
            'solid_b': {'arg_a': DependencyDefinition('solid_a')},
            'solid_c': {'arg_a': DependencyDefinition('solid_a')},
            'solid_d': {
                'arg_b': DependencyDefinition('solid_b'),
                'arg_c': DependencyDefinition('solid_c'),
            },
        },
    )



import six

from dagster import check
from dagster.core.definitions import (
    Materialization,
    OutputDefinition,
    PipelineDefinition,
    Solid,
    SolidHandle,
)

from dagster.core.errors import DagsterInvariantViolationError
from .objects import (
    ExecutionStep,
    ExecutionValueSubplan,
    StepInput,
    StepKind,
    StepOutput,
    StepOutputValue,
)

from .utility import create_joining_subplan

MATERIALIZATION_THUNK_INPUT = 'materialization_thunk_input'
MATERIALIZATION_THUNK_OUTPUT = 'materialization_thunk_output'


def _create_materialization_lambda(output_def, config_spec):
    check.inst_param(output_def, 'output_def', OutputDefinition)
    check.invariant(output_def.runtime_type.output_schema, 'Must have output schema')

    def _fn(step_context, inputs):
        runtime_value = inputs[MATERIALIZATION_THUNK_INPUT]
        path = output_def.runtime_type.output_schema.materialize_runtime_value(
            step_context, config_spec, runtime_value
        )

        if not isinstance(path, six.string_types):
            raise DagsterInvariantViolationError(
                (
                    'materialize_runtime_value on type {type_name} has returned '
                    'value {value} of type {python_type}. You must return a '
                    'string (and ideally a valid file path).'
                ).format(
                    type_name=output_def.runtime_type.name,
                    value=repr(path),
                    python_type=type(path).__name__,
                )
            )

        yield StepOutputValue(output_name=MATERIALIZATION_THUNK_OUTPUT, value=runtime_value)
        yield Materialization(
            path=path,
            description=('Materialization of {solid_name}.{output_name}').format(
                output_name=output_def.name, solid_name=step_context.solid_handle.name
            ),
        )

    return _fn


def configs_for_output(solid, solid_config, output_def):
    for output_spec in solid_config.outputs:
        check.invariant(len(output_spec) == 1)
        output_name, output_spec = list(output_spec.items())[0]
        check.invariant(solid.has_output(output_name))
        if output_name == output_def.name:
            yield output_spec


def decorate_with_output_materializations(
    pipeline_def, environment_config, solid, output_def, subplan
):
    check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition)
    check.inst_param(solid, 'solid', Solid)
    check.inst_param(output_def, 'output_def', OutputDefinition)
    check.inst_param(subplan, 'subplan', ExecutionValueSubplan)

    solid_config = environment_config.solids.get(solid.name)

    if not (solid_config and solid_config.outputs):
        return subplan

    new_steps = []

    for mat_count, output_spec in enumerate(configs_for_output(solid, solid_config, output_def)):
        new_steps.append(
            ExecutionStep(
                pipeline_name=pipeline_def.name,
                key_suffix='outputs.{output}.materialize.{mat_count}'.format(
                    output=output_def.name, mat_count=mat_count
                ),
                step_inputs=[
                    StepInput(
                        name=MATERIALIZATION_THUNK_INPUT,
                        runtime_type=output_def.runtime_type,
                        prev_output_handle=subplan.terminal_step_output_handle,
                    )
                ],
                step_outputs=[
                    StepOutput(
                        name=MATERIALIZATION_THUNK_OUTPUT,
                        runtime_type=output_def.runtime_type,
                        optional=output_def.optional,
                    )
                ],
                kind=StepKind.MATERIALIZATION_THUNK,
                solid_handle=SolidHandle(solid.name, solid.definition.name),
                compute_fn=_create_materialization_lambda(output_def, output_spec),
            )
        )

    return create_joining_subplan(
        pipeline_def,
        solid,
        'outputs.{output}.materialize.join'.format(output=output_def.name),
        new_steps,
        MATERIALIZATION_THUNK_OUTPUT,
    )



import tempfile
import os

import pytest

from dagster.core.types.marshal import (
    PickleSerializationStrategy,
    serialize_to_file,
    deserialize_from_file,
)
from dagster.utils.test import yield_empty_pipeline_context


# https://dev.azure.com/elementl/dagster/_build/results?buildId=2941
@pytest.mark.skipif(
    os.name == 'nt', reason='Azure pipelines does not let us use tempfile.NamedTemporaryFile'
)
def test_serialize_deserialize():
    with yield_empty_pipeline_context() as context:
        with tempfile.NamedTemporaryFile() as fd:
            serialize_to_file(context, PickleSerializationStrategy(), 'foo', fd.name)
            assert deserialize_from_file(context, PickleSerializationStrategy(), fd.name) == 'foo'



from collections import defaultdict

import pytest

from dagster import (
    DagsterInvalidDefinitionError,
    DagsterInvariantViolationError,
    PipelineDefinition,
    RepositoryDefinition,
    SolidDefinition,
    lambda_solid,
    solid,
    Field,
    Bool,
    Dict,
    execute_pipeline,
    DagsterExecutionStepExecutionError,
)
from dagster.utils import script_relative_path


def create_single_node_pipeline(name, called):
    called[name] = called[name] + 1
    return PipelineDefinition(
        name=name,
        solids=[
            SolidDefinition(
                name=name + '_solid',
                inputs=[],
                outputs=[],
                transform_fn=lambda *_args, **_kwargs: None,
            )
        ],
    )


def test_repo_definition():
    called = defaultdict(int)
    repo = RepositoryDefinition(
        name='some_repo',
        pipeline_dict={
            'foo': lambda: create_single_node_pipeline('foo', called),
            'bar': lambda: create_single_node_pipeline('bar', called),
        },
    )

    foo_pipeline = repo.get_pipeline('foo')
    assert isinstance(foo_pipeline, PipelineDefinition)
    assert foo_pipeline.name == 'foo'

    assert 'foo' in called
    assert called['foo'] == 1
    assert 'bar' not in called

    bar_pipeline = repo.get_pipeline('bar')
    assert isinstance(bar_pipeline, PipelineDefinition)
    assert bar_pipeline.name == 'bar'

    assert 'foo' in called
    assert called['foo'] == 1
    assert 'bar' in called
    assert called['bar'] == 1

    foo_pipeline = repo.get_pipeline('foo')
    assert isinstance(foo_pipeline, PipelineDefinition)
    assert foo_pipeline.name == 'foo'

    assert 'foo' in called
    assert called['foo'] == 1

    pipelines = repo.get_all_pipelines()

    assert set(['foo', 'bar']) == {pipeline.name for pipeline in pipelines}

    assert repo.get_solid_def('foo_solid').name == 'foo_solid'
    assert repo.get_solid_def('bar_solid').name == 'bar_solid'


def test_dupe_solid_repo_definition():
    @lambda_solid(name='same')
    def noop():
        pass

    @lambda_solid(name='same')
    def noop2():
        pass

    repo = RepositoryDefinition(
        'error_repo',
        pipeline_dict={
            'first': lambda: PipelineDefinition(name='first', solids=[noop]),
            'second': lambda: PipelineDefinition(name='second', solids=[noop2]),
        },
    )

    with pytest.raises(DagsterInvalidDefinitionError) as exc_info:
        repo.get_all_pipelines()

    assert str(exc_info.value) == (
        'You have defined two solids named "same" in repository "error_repo". '
        'Solid names must be unique within a repository. The solid has been defined '
        'in pipeline "first" and it has been defined again in pipeline "second."'
    )


def test_dupe_solid_repo_definition_error_opt_out():
    @lambda_solid(name='same')
    def noop():
        pass

    @lambda_solid(name='same')
    def noop2():
        pass

    repo = RepositoryDefinition(
        'skip_error_repo',
        pipeline_dict={
            'first': lambda: PipelineDefinition(name='first', solids=[noop]),
            'second': lambda: PipelineDefinition(name='second', solids=[noop2]),
        },
        enforce_solid_def_uniqueness=False,
    )

    assert repo.get_all_pipelines()

    with pytest.raises(DagsterInvariantViolationError):
        repo.get_solid_def('foo')


def test_repo_config():
    @solid(config_field=Field(Dict(fields={'error': Field(Bool)})))
    def can_fail(context):
        if context.solid_config['error']:
            raise Exception('I did an error')
        return 'cool'

    @lambda_solid
    def always_fail():
        raise Exception('I always do this')

    repo = RepositoryDefinition(
        'config_test',
        pipeline_dict={
            'simple': lambda: PipelineDefinition(name='simple', solids=[can_fail, always_fail])
        },
        repo_config={
            'pipelines': {
                'simple': {
                    'presets': {
                        'passing': {
                            'environment_files': [script_relative_path('pass_env.yml')],
                            'solid_subset': ['can_fail'],
                        },
                        'failing_1': {
                            'environment_files': [script_relative_path('fail_env.yml')],
                            'solid_subset': ['can_fail'],
                        },
                        'failing_2': {'environment_files': [script_relative_path('pass_env.yml')]},
                        'invalid_1': {
                            'environment_files': [script_relative_path('not_a_file.yml')]
                        },
                        'invalid_2': {
                            'environment_files': [
                                script_relative_path('test_repository_definition.py')
                            ]
                        },
                    }
                }
            }
        },
    )

    execute_pipeline(**repo.get_preset_pipeline('simple', 'passing'))

    with pytest.raises(DagsterExecutionStepExecutionError):
        execute_pipeline(**repo.get_preset_pipeline('simple', 'failing_1'))

    with pytest.raises(DagsterExecutionStepExecutionError):
        execute_pipeline(**repo.get_preset_pipeline('simple', 'failing_2'))

    with pytest.raises(DagsterInvalidDefinitionError, match="not_a_file.yml"):
        execute_pipeline(**repo.get_preset_pipeline('simple', 'invalid_1'))

    with pytest.raises(DagsterInvariantViolationError, match="error attempting to parse yaml"):
        execute_pipeline(**repo.get_preset_pipeline('simple', 'invalid_2'))

    with pytest.raises(DagsterInvariantViolationError, match="Could not find pipeline"):
        execute_pipeline(**repo.get_preset_pipeline('not_real', 'passing'))

    with pytest.raises(DagsterInvariantViolationError, match="Could not find preset"):
        execute_pipeline(**repo.get_preset_pipeline('simple', 'not_failing'))



'''
This module defines the different notions of context passed to user code during pipeline
execution. There are currently four types of contexts defined:

1) StepExecutionContext
2) TransformExecutionContext
3) ExpectationExecutionContext

They inherit from each other in reverse order. They are also relatively lightweight
objects, and a new one is created for every invocation into user space.
'''

from abc import ABCMeta, abstractproperty, abstractmethod
from collections import namedtuple
import logging
import six
from dagster import check
from dagster.utils.logging import INFO, define_colored_console_logger

from .execution_context import (
    SystemExpectationExecutionContext,
    SystemStepExecutionContext,
    SystemTransformExecutionContext,
)


class ExecutionContext(namedtuple('_ExecutionContext', 'loggers resources tags')):
    '''
    The user-facing object in the context creation function. The user constructs
    this in order to effect the context creation process. This could be named
    SystemPipelineExecutionContextCreationData although that seemed excessively verbose.

    Args:
        loggers (List[Logger]):
        resources ():
        tags (dict[str, str])
    '''

    def __new__(cls, loggers=None, resources=None, tags=None):
        return super(ExecutionContext, cls).__new__(
            cls,
            loggers=check.opt_list_param(loggers, 'loggers', logging.Logger),
            resources=resources,
            tags=check.opt_dict_param(tags, 'tags'),
        )

    @staticmethod
    def console_logging(log_level=INFO, resources=None):
        return ExecutionContext(
            loggers=[define_colored_console_logger('dagster', log_level)], resources=resources
        )


class StepExecutionContext(object):
    __slots__ = ['_system_step_execution_context', '_legacy_context']

    def __init__(self, system_step_execution_context):
        self._system_step_execution_context = check.inst_param(
            system_step_execution_context,
            'system_step_execution_context',
            SystemStepExecutionContext,
        )

    @property
    def resources(self):
        return self._system_step_execution_context.resources

    @property
    def run_id(self):
        return self._system_step_execution_context.run_id

    @property
    def environment_dict(self):
        return self._system_step_execution_context.environment_dict

    @property
    def pipeline_def(self):
        return self._system_step_execution_context.pipeline_def

    @property
    def log(self):
        return self._system_step_execution_context.log

    @property
    def solid_handle(self):
        return self._system_step_execution_context.solid_handle

    @property
    def solid(self):
        return self._system_step_execution_context.pipeline_def.get_solid(self.solid_handle)

    @property
    def solid_def(self):
        return self._system_step_execution_context.pipeline_def.get_solid(
            self.solid_handle
        ).definition

    def has_tag(self, key):
        return self._system_step_execution_context.has_tag(key)

    def get_tag(self, key):
        return self._system_step_execution_context.get_tag(key)

    def get_system_context(self):
        '''
        This allows advanced users (e.g. framework authors) to punch through
        to the underlying system context.
        '''
        return self._system_step_execution_context


class AbstractTransformExecutionContext(six.with_metaclass(ABCMeta)):  # pylint: disable=no-init
    @abstractmethod
    def has_tag(self, key):
        pass

    @abstractmethod
    def get_tag(self, key):
        pass

    @abstractproperty
    def run_id(self):
        pass

    @abstractproperty
    def solid_def(self):
        pass

    @abstractproperty
    def solid(self):
        pass

    @abstractproperty
    def pipeline_def(self):
        pass

    @abstractproperty
    def resources(self):
        pass

    @abstractproperty
    def log(self):
        pass


class TransformExecutionContext(StepExecutionContext, AbstractTransformExecutionContext):
    __slots__ = ['_system_transform_execution_context']

    def __init__(self, system_transform_execution_context):
        self._system_transform_execution_context = check.inst_param(
            system_transform_execution_context,
            'system_transform_execution_context',
            SystemTransformExecutionContext,
        )
        super(TransformExecutionContext, self).__init__(system_transform_execution_context)

    @property
    def solid_config(self):
        return self._system_transform_execution_context.solid_config


class ExpectationExecutionContext(StepExecutionContext):
    __slots__ = ['_system_expectation_execution_context']

    def __init__(self, system_expectation_execution_context):
        self._system_expectation_execution_context = check.inst_param(
            system_expectation_execution_context,
            'system_expectation_execution_context',
            SystemExpectationExecutionContext,
        )
        super(ExpectationExecutionContext, self).__init__(system_expectation_execution_context)

    @property
    def expectation_def(self):
        return self._system_expectation_execution_context.expectation_def

    @property
    def inout_def(self):
        return self._system_expectation_execution_context.inout_def



import runpy
import sys
import warnings

from dagster import execute_pipeline, RunConfig
from dagster.tutorials.intro_tutorial.resources import define_resource_test_pipeline


def has_message(events, message):
    for event in events:
        if message == event.user_message:
            return True

    return False


def test_run_local():
    events = []

    def _event_callback(ev):
        events.append(ev)

    result = execute_pipeline(
        define_resource_test_pipeline(),
        environment_dict={
            'context': {'local': {}},
            'solids': {'add_ints': {'inputs': {'num_one': {'value': 2}, 'num_two': {'value': 3}}}},
        },
        run_config=RunConfig(event_callback=_event_callback),
    )

    assert result.success
    assert result.result_for_solid('add_ints').transformed_value() == 5

    assert has_message(events, 'Setting key=add value=5 in memory')
    assert not has_message(events, 'Setting key=add value=5 in cloud')


def test_run_cloud():
    events = []

    def _event_callback(ev):
        events.append(ev)

    result = execute_pipeline(
        define_resource_test_pipeline(),
        environment_dict={
            'context': {
                'cloud': {
                    'resources': {
                        'store': {'config': {'username': 'some_user', 'password': 'some_password'}}
                    }
                }
            },
            'solids': {'add_ints': {'inputs': {'num_one': {'value': 2}, 'num_two': {'value': 6}}}},
        },
        run_config=RunConfig(event_callback=_event_callback),
    )

    assert result.success
    assert result.result_for_solid('add_ints').transformed_value() == 8

    assert not has_message(events, 'Setting key=add value=8 in memory')
    assert has_message(events, 'Setting key=add value=8 in cloud')


def test_resources():
    # Why?!? Because
    # http://python-notes.curiousefficiency.org/en/latest/python_concepts/import_traps.html#the-double-import-trap
    # (Nothing to see here)
    if not sys.warnoptions:  # allow overriding with `-W` option
        warnings.filterwarnings('ignore', category=RuntimeWarning, module='runpy')

    runpy.run_module('dagster.tutorials.intro_tutorial.resources', run_name='__main__')



import pytest

from dagster import (
    DependencyDefinition,
    Dict,
    Field,
    InputDefinition,
    OutputDefinition,
    PipelineContextDefinition,
    PipelineDefinition,
    SolidInstance,
    String,
    lambda_solid,
    solid,
    types,
)
from dagster.core.definitions import Solid, Materialization, create_environment_schema
from dagster.core.definitions.dependency import SolidOutputHandle
from dagster.core.errors import DagsterInvalidDefinitionError


def test_deps_equal():
    assert DependencyDefinition('foo') == DependencyDefinition('foo')
    assert DependencyDefinition('foo') != DependencyDefinition('bar')

    assert DependencyDefinition('foo', 'bar') == DependencyDefinition('foo', 'bar')
    assert DependencyDefinition('foo', 'bar') != DependencyDefinition('foo', 'quuz')


def test_solid_def():
    @lambda_solid
    def produce_string():
        return 'foo'

    @solid(
        inputs=[InputDefinition('input_one', types.String)],
        outputs=[OutputDefinition(types.Any)],
        config_field=Field(Dict({'another_field': Field(types.Int)})),
    )
    def solid_one(_context, input_one):
        raise Exception('should not execute')

    pipeline_def = PipelineDefinition(
        solids=[produce_string, solid_one],
        dependencies={'solid_one': {'input_one': DependencyDefinition('produce_string')}},
    )

    assert len(pipeline_def.solids[0].output_handles()) == 1

    assert isinstance(pipeline_def.solid_named('solid_one'), Solid)

    solid_one_solid = pipeline_def.solid_named('solid_one')

    assert solid_one_solid.has_input('input_one')

    assert isinstance(solid_one_solid.input_def_named('input_one'), InputDefinition)

    assert len(solid_one_solid.input_dict) == 1
    assert len(solid_one_solid.output_dict) == 1

    assert str(solid_one_solid.input_handle('input_one')) == (
        'SolidInputHandle(definition_name="\'solid_one\'", input_name="\'input_one\'", '
        'solid_name="\'solid_one\'")'
    )

    assert repr(solid_one_solid.input_handle('input_one')) == (
        'SolidInputHandle(definition_name="\'solid_one\'", input_name="\'input_one\'", '
        'solid_name="\'solid_one\'")'
    )

    assert str(solid_one_solid.output_handle('result')) == (
        'SolidOutputHandle(definition_name="\'solid_one\'", output_name="\'result\'", '
        'solid_name="\'solid_one\'")'
    )

    assert repr(solid_one_solid.output_handle('result')) == (
        'SolidOutputHandle(definition_name="\'solid_one\'", output_name="\'result\'", '
        'solid_name="\'solid_one\'")'
    )

    assert solid_one_solid.output_handle('result') == SolidOutputHandle(
        solid_one_solid, solid_one_solid.output_dict['result']
    )

    assert len(pipeline_def.dependency_structure.deps_of_solid_with_input('solid_one')) == 1

    assert len(pipeline_def.dependency_structure.depended_by_of_solid('produce_string')) == 1

    assert len(pipeline_def.dependency_structure.input_handles()) == 1

    assert len(pipeline_def.dependency_structure.items()) == 1


def test_pipeline_types():
    @lambda_solid
    def produce_string():
        return 'foo'

    @solid(
        inputs=[InputDefinition('input_one', types.String)],
        outputs=[OutputDefinition(types.Any)],
        config_field=Field(Dict({'another_field': Field(types.Int)})),
    )
    def solid_one(_context, input_one):
        raise Exception('should not execute')

    pipeline_def = PipelineDefinition(
        solids=[produce_string, solid_one],
        dependencies={'solid_one': {'input_one': DependencyDefinition('produce_string')}},
        context_definitions={
            'context_one': PipelineContextDefinition(
                context_fn=lambda: None, config_field=Field(Dict({'field_one': Field(String)}))
            )
        },
    )

    environment_schema = create_environment_schema(pipeline_def)

    assert environment_schema.has_config_type('String')
    assert environment_schema.has_config_type('Int')
    assert not environment_schema.has_config_type('SomeName')


def test_mapper_errors():
    @lambda_solid
    def solid_a():
        print('a: 1')
        return 1

    with pytest.raises(DagsterInvalidDefinitionError) as excinfo_1:
        PipelineDefinition(
            solids=[solid_a], dependencies={'solid_b': {'arg_a': DependencyDefinition('solid_a')}}
        )
    assert str(excinfo_1.value) == 'Solid solid_b in dependency dictionary not found in solid list'

    with pytest.raises(DagsterInvalidDefinitionError) as excinfo_2:
        PipelineDefinition(
            solids=[solid_a],
            dependencies={
                SolidInstance('solid_b', alias='solid_c'): {
                    'arg_a': DependencyDefinition('solid_a')
                }
            },
        )
    assert (
        str(excinfo_2.value)
        == 'Solid solid_b (aliased by solid_c in dependency dictionary) not found in solid list'
    )


def test_materialization():
    assert isinstance(Materialization('foo', 'foo.txt'), Materialization)



from dagster import (
    DependencyDefinition,
    InputDefinition,
    OutputDefinition,
    PipelineDefinition,
    SolidInstance,
    execute_solid,
    execute_solids,
    lambda_solid,
    Int,
)


@lambda_solid(
    inputs=[InputDefinition('num1', Int), InputDefinition('num2', Int)],
    output=OutputDefinition(Int),
)
def adder(num1, num2):
    return num1 + num2


@lambda_solid(
    inputs=[InputDefinition('num1', Int), InputDefinition('num2', Int)],
    output=OutputDefinition(Int),
)
def multer(num1, num2):
    return num1 * num2


def define_part_fourteen_step_one_pipeline():
    # (a + b) * (c + d)

    return PipelineDefinition(
        name='part_fourteen_step_one_pipeline',
        solids=[adder, multer],
        dependencies={
            SolidInstance(adder.name, 'a_plus_b'): {},
            SolidInstance(adder.name, 'c_plus_d'): {},
            SolidInstance(multer.name, 'final'): {
                'num1': DependencyDefinition('a_plus_b'),
                'num2': DependencyDefinition('c_plus_d'),
            },
        },
    )


def execute_test_only_final():
    solid_result = execute_solid(
        define_part_fourteen_step_one_pipeline(), 'final', inputs={'num1': 3, 'num2': 4}
    )
    assert solid_result.success
    assert solid_result.transformed_value() == 12


def execute_test_a_plus_b_final_subdag():
    results = execute_solids(
        define_part_fourteen_step_one_pipeline(),
        ['a_plus_b', 'final'],
        inputs={'a_plus_b': {'num1': 2, 'num2': 4}, 'final': {'num2': 6}},
    )

    assert results['a_plus_b'].transformed_value() == 6
    assert results['final'].transformed_value() == 36



from __future__ import print_function

import os

import pytest

from click.testing import CliRunner

from dagster.core.runs import base_run_directory
from dagster.cli.pipeline import (
    execute_print_command,
    execute_list_command,
    execute_execute_command,
    execute_scaffold_command,
    pipeline_execute_command,
    pipeline_list_command,
    pipeline_print_command,
    pipeline_scaffold_command,
)
from dagster.cli.dynamic_loader import InvalidRepositoryLoadingComboError

from dagster.utils import script_relative_path


def no_print(_):
    return None


def test_list_command():
    runner = CliRunner()

    execute_list_command(
        {
            'repository_yaml': None,
            'python_file': script_relative_path('test_dynamic_loader.py'),
            'module_name': None,
            'fn_name': 'define_bar_repo',
        },
        no_print,
    )

    result = runner.invoke(
        pipeline_list_command,
        ['-f', script_relative_path('test_dynamic_loader.py'), '-n', 'define_bar_repo'],
    )
    assert result.exit_code == 0
    assert result.output == (
        'Repository bar\n'
        '**************\n'
        'Pipeline: foo\n'
        'Solids: (Execution Order)\n'
        '    do_something\n'
    )

    execute_list_command(
        {
            'repository_yaml': None,
            'python_file': None,
            'module_name': 'dagster.tutorials.intro_tutorial.repos',
            'fn_name': 'define_repo',
        },
        no_print,
    )

    result = runner.invoke(
        pipeline_list_command, ['-m', 'dagster.tutorials.intro_tutorial.repos', '-n', 'define_repo']
    )
    assert result.exit_code == 0
    assert result.output == (
        'Repository demo_repository\n'
        '**************************\n'
        'Pipeline: repo_demo_pipeline\n'
        'Solids: (Execution Order)\n'
        '    hello_world\n'
    )

    execute_list_command(
        {
            'repository_yaml': script_relative_path('repository.yml'),
            'python_file': None,
            'module_name': None,
            'fn_name': None,
        },
        no_print,
    )

    result = runner.invoke(pipeline_list_command, ['-y', script_relative_path('repository.yml')])
    assert result.exit_code == 0
    assert result.output == (
        'Repository demo_repository\n'
        '**************************\n'
        'Pipeline: repo_demo_pipeline\n'
        'Solids: (Execution Order)\n'
        '    hello_world\n'
    )

    with pytest.raises(InvalidRepositoryLoadingComboError):
        execute_list_command(
            {
                'repository_yaml': None,
                'python_file': 'foo.py',
                'module_name': 'dagster.tutorials.intro_tutorial.repos',
                'fn_name': 'define_repo',
            },
            no_print,
        )

    result = runner.invoke(
        pipeline_list_command,
        ['-f', 'foo.py', '-m', 'dagster.tutorials.intro_tutorial.repos', '-n', 'define_repo'],
    )
    assert result.exit_code == 1
    assert isinstance(result.exception, InvalidRepositoryLoadingComboError)

    with pytest.raises(InvalidRepositoryLoadingComboError):
        execute_list_command(
            {
                'repository_yaml': None,
                'python_file': None,
                'module_name': 'dagster.tutorials.intro_tutorial.repos',
                'fn_name': None,
            },
            no_print,
        )

    result = runner.invoke(pipeline_list_command, ['-m', 'dagster.tutorials.intro_tutorial.repos'])
    assert result.exit_code == 1
    assert isinstance(result.exception, InvalidRepositoryLoadingComboError)

    with pytest.raises(InvalidRepositoryLoadingComboError):
        execute_list_command(
            {
                'repository_yaml': None,
                'python_file': script_relative_path('test_dynamic_loader.py'),
                'module_name': None,
                'fn_name': None,
            },
            no_print,
        )

    result = runner.invoke(
        pipeline_list_command, ['-f', script_relative_path('test_dynamic_loader.py')]
    )
    assert result.exit_code == 1
    assert isinstance(result.exception, InvalidRepositoryLoadingComboError)


def valid_execute_args():
    return [
        {
            'repository_yaml': script_relative_path('repository_file.yml'),
            'pipeline_name': ('foo',),
            'python_file': None,
            'module_name': None,
            'fn_name': None,
        },
        {
            'repository_yaml': script_relative_path('repository_module.yml'),
            'pipeline_name': ('repo_demo_pipeline',),
            'python_file': None,
            'module_name': None,
            'fn_name': None,
        },
        {
            'repository_yaml': None,
            'pipeline_name': ('foo',),
            'python_file': script_relative_path('test_dynamic_loader.py'),
            'module_name': None,
            'fn_name': 'define_bar_repo',
        },
        {
            'repository_yaml': None,
            'pipeline_name': ('repo_demo_pipeline',),
            'python_file': None,
            'module_name': 'dagster.tutorials.intro_tutorial.repos',
            'fn_name': 'define_repo',
        },
        {
            'repository_yaml': None,
            'pipeline_name': (),
            'python_file': None,
            'module_name': 'dagster.tutorials.intro_tutorial.repos',
            'fn_name': 'define_repo_demo_pipeline',
        },
        {
            'repository_yaml': None,
            'pipeline_name': (),
            'python_file': script_relative_path('test_dynamic_loader.py'),
            'module_name': None,
            'fn_name': 'define_foo_pipeline',
        },
    ]


def valid_cli_args():
    return [
        ['-y', script_relative_path('repository_file.yml'), 'foo'],
        ['-y', script_relative_path('repository_module.yml'), 'repo_demo_pipeline'],
        ['-f', script_relative_path('test_dynamic_loader.py'), '-n', 'define_bar_repo', 'foo'],
        ['-m', 'dagster.tutorials.intro_tutorial.repos', '-n', 'define_repo', 'repo_demo_pipeline'],
        ['-m', 'dagster.tutorials.intro_tutorial.repos', '-n', 'define_repo_demo_pipeline'],
        ['-f', script_relative_path('test_dynamic_loader.py'), '-n', 'define_foo_pipeline'],
    ]


def test_print_command():
    for cli_args in valid_execute_args():
        execute_print_command(verbose=True, cli_args=cli_args, print_fn=no_print)

    for cli_args in valid_execute_args():
        execute_print_command(verbose=False, cli_args=cli_args, print_fn=no_print)

    runner = CliRunner()

    for cli_args in valid_cli_args():
        result = runner.invoke(pipeline_print_command, cli_args)
        assert result.exit_code == 0

        result = runner.invoke(pipeline_print_command, ['--verbose'] + cli_args)
        assert result.exit_code == 0


def test_execute_command():
    for cli_args in valid_execute_args():
        execute_execute_command(env=None, raise_on_error=True, cli_args=cli_args)

    for cli_args in valid_execute_args():
        execute_execute_command(
            env=[script_relative_path('env.yml')], raise_on_error=True, cli_args=cli_args
        )

    runner = CliRunner()

    for cli_args in valid_cli_args():
        result = runner.invoke(pipeline_execute_command, cli_args)
        assert result.exit_code == 0

        result = runner.invoke(
            pipeline_execute_command, ['--env', script_relative_path('env.yml')] + cli_args
        )
        assert result.exit_code == 0


def test_scaffold_command():
    for cli_args in valid_execute_args():
        cli_args['print_only_required'] = True
        execute_scaffold_command(cli_args=cli_args, print_fn=no_print)

        cli_args['print_only_required'] = False
        execute_scaffold_command(cli_args=cli_args, print_fn=no_print)

    runner = CliRunner()

    for cli_args in valid_cli_args():
        result = runner.invoke(pipeline_scaffold_command, cli_args)
        assert result.exit_code == 0

        result = runner.invoke(pipeline_scaffold_command, ['-p'] + cli_args)
        assert result.exit_code == 0


def test_default_memory_run_storage():
    cli_args = {
        'repository_yaml': script_relative_path('repository_file.yml'),
        'pipeline_name': ('foo',),
        'python_file': None,
        'module_name': None,
        'fn_name': None,
    }
    result = execute_execute_command(env=None, raise_on_error=True, cli_args=cli_args)
    assert result.success

    run_dir = os.path.join(base_run_directory(), result.run_id)

    assert not os.path.isdir(run_dir)


def test_override_with_in_memory_storage():
    cli_args = {
        'repository_yaml': script_relative_path('repository_file.yml'),
        'pipeline_name': ('foo',),
        'python_file': None,
        'module_name': None,
        'fn_name': None,
    }
    result = execute_execute_command(
        env=[script_relative_path('in_memory_env.yml')], raise_on_error=True, cli_args=cli_args
    )
    assert result.success

    run_dir = os.path.join(base_run_directory(), result.run_id)

    assert not os.path.exists(run_dir)


def test_override_with_filesystem_storage():
    cli_args = {
        'repository_yaml': script_relative_path('repository_file.yml'),
        'pipeline_name': ('foo',),
        'python_file': None,
        'module_name': None,
        'fn_name': None,
    }
    result = execute_execute_command(
        env=[script_relative_path('filesystem_env.yml')], raise_on_error=True, cli_args=cli_args
    )
    assert result.success

    run_dir = os.path.join(base_run_directory(), result.run_id)

    assert os.path.exists(run_dir)



from dagster import check
from dagster.core.definitions import (
    ExpectationResult,
    Materialization,
    PipelineDefinition,
    Result,
    Solid,
    SolidHandle,
)
from dagster.core.errors import DagsterInvariantViolationError
from dagster.core.execution.execution_context import SystemTransformExecutionContext
from dagster.core.execution.user_context import TransformExecutionContext

from .objects import ExecutionStep, StepInput, StepKind, StepOutput, StepOutputValue


def create_transform_step(pipeline_def, environment_config, solid, step_inputs):
    check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition)
    check.inst_param(solid, 'solid', Solid)
    check.list_param(step_inputs, 'step_inputs', of_type=StepInput)

    return ExecutionStep(
        pipeline_name=pipeline_def.name,
        key_suffix='transform',
        step_inputs=step_inputs,
        step_outputs=[
            StepOutput(
                name=name, runtime_type=output_def.runtime_type, optional=output_def.optional
            )
            for name, output_def in solid.definition.output_dict.items()
        ],
        compute_fn=lambda step_context, inputs: _execute_core_transform(
            step_context.for_transform(), inputs, solid.definition.transform_fn
        ),
        kind=StepKind.TRANSFORM,
        solid_handle=SolidHandle(solid.name, solid.definition.name),
        metadata=solid.step_metadata_fn(environment_config) if solid.step_metadata_fn else {},
    )


def _yield_transform_results(transform_context, inputs, transform_fn):
    check.inst_param(transform_context, 'transform_context', SystemTransformExecutionContext)
    step = transform_context.step
    gen = transform_fn(TransformExecutionContext(transform_context), inputs)

    if isinstance(gen, Result):
        raise DagsterInvariantViolationError(
            (
                'Transform for solid {solid_name} returned a Result rather than '
                'yielding it. The transform_fn of the core SolidDefinition must yield '
                'its results'
            ).format(solid_name=step.solid_handle.name)
        )

    if gen is None:
        return

    for result in gen:
        if isinstance(result, Result):
            transform_context.log.info(
                'Solid {solid} emitted output "{output}" value {value}'.format(
                    solid=step.solid_handle.name,
                    output=result.output_name,
                    value=repr(result.value),
                )
            )
            yield StepOutputValue(output_name=result.output_name, value=result.value)

        elif isinstance(result, (Materialization, ExpectationResult)):
            yield result

        else:
            raise DagsterInvariantViolationError(
                (
                    'Transform for solid {solid_name} yielded {result} rather an '
                    'an instance of the Result or Materialization class.'
                ).format(result=repr(result), solid_name=step.solid_handle.name)
            )


def _execute_core_transform(transform_context, inputs, transform_fn):
    '''
    Execute the user-specified transform for the solid. Wrap in an error boundary and do
    all relevant logging and metrics tracking
    '''
    check.inst_param(transform_context, 'transform_context', SystemTransformExecutionContext)
    check.dict_param(inputs, 'inputs', key_type=str)

    step = transform_context.step

    transform_context.log.debug(
        'Executing core transform for solid {solid}.'.format(solid=step.solid_handle.name)
    )

    all_results = []
    for step_output in _yield_transform_results(transform_context, inputs, transform_fn):
        yield step_output
        if isinstance(step_output, StepOutputValue):
            all_results.append(step_output)

    if len(all_results) != len(step.step_outputs):
        emitted_result_names = {r.output_name for r in all_results}
        solid_output_names = {output.name for output in step.step_outputs}
        omitted_outputs = solid_output_names.difference(emitted_result_names)
        transform_context.log.info(
            'Solid {solid} did not fire outputs {outputs}'.format(
                solid=step.solid_handle.name, outputs=repr(omitted_outputs)
            )
        )



from dagster import (
    ExpectationDefinition,
    ExpectationResult,
    InputDefinition,
    Int,
    OutputDefinition,
    PipelineDefinition,
    lambda_solid,
)


@lambda_solid(
    inputs=[
        InputDefinition(
            'num_one',
            Int,
            expectations=[
                ExpectationDefinition(
                    name='check_positive',
                    expectation_fn=lambda _info, value: ExpectationResult(success=value > 0),
                )
            ],
        ),
        InputDefinition('num_two', Int),
    ],
    output=OutputDefinition(Int),
)
def add_ints(num_one, num_two):
    return num_one + num_two


def define_expectations_tutorial_pipeline():
    return PipelineDefinition(name='expectations_tutorial_pipeline', solids=[add_ints])



import pytest

from dagster import (
    lambda_solid,
    solid,
    InputDefinition,
    OutputDefinition,
    Nothing,
    Int,
    PipelineDefinition,
    execute_pipeline,
    Result,
    DependencyDefinition,
    DagsterInvalidDefinitionError,
    DagsterInvariantViolationError,
    SolidInstance,
)


def _define_nothing_dep_pipeline():
    @lambda_solid(output=OutputDefinition(Nothing, 'complete'))
    def start_nothing():
        pass

    @lambda_solid(
        inputs=[
            InputDefinition('add_complete', Nothing),
            InputDefinition('yield_complete', Nothing),
        ]
    )
    def end_nothing():
        pass

    @lambda_solid(output=OutputDefinition(Int))
    def emit_value():
        return 1

    @lambda_solid(
        inputs=[InputDefinition('on_complete', Nothing), InputDefinition('num', Int)],
        output=OutputDefinition(Int),
    )
    def add_value(num):
        return 1 + num

    @solid(
        name='yield_values',
        inputs=[InputDefinition('on_complete', Nothing)],
        outputs=[
            OutputDefinition(Int, 'num_1'),
            OutputDefinition(Int, 'num_2'),
            OutputDefinition(Nothing, 'complete'),
        ],
    )
    def yield_values(_context):
        yield Result(1, 'num_1')
        yield Result(2, 'num_2')
        yield Result(None, 'complete')

    return PipelineDefinition(
        name='simple_exc',
        solids=[emit_value, add_value, start_nothing, end_nothing, yield_values],
        dependencies={
            'add_value': {
                'on_complete': DependencyDefinition('start_nothing', 'complete'),
                'num': DependencyDefinition('emit_value'),
            },
            'yield_values': {'on_complete': DependencyDefinition('start_nothing', 'complete')},
            'end_nothing': {
                'add_complete': DependencyDefinition('add_value'),
                'yield_complete': DependencyDefinition('yield_values', 'complete'),
            },
        },
    )


def test_valid_nothing_dependencies():

    result = execute_pipeline(_define_nothing_dep_pipeline())

    assert result.success


def test_invalid_input_dependency():
    @lambda_solid(output=OutputDefinition(Nothing))
    def do_nothing():
        pass

    @lambda_solid(inputs=[InputDefinition('num', Int)], output=OutputDefinition(Int))
    def add_one(num):
        return num + 1

    with pytest.raises(DagsterInvalidDefinitionError):
        PipelineDefinition(
            name='bad_dep',
            solids=[do_nothing, add_one],
            dependencies={'add_one': {'num': DependencyDefinition('do_nothing')}},
        )


def test_result_type_check():
    @solid(outputs=[OutputDefinition(Nothing)])
    def bad(_context):
        yield Result('oops')

    pipeline = PipelineDefinition(name='fail', solids=[bad])
    with pytest.raises(DagsterInvariantViolationError):
        execute_pipeline(pipeline)


def test_nothing_inputs():
    @lambda_solid(inputs=[InputDefinition('never_defined', Nothing)])
    def emit_one():
        return 1

    @lambda_solid
    def emit_two():
        return 2

    @lambda_solid
    def emit_three():
        return 3

    @lambda_solid(output=OutputDefinition(Nothing))
    def emit_nothing():
        pass

    @solid(
        inputs=[
            InputDefinition('_one', Nothing),
            InputDefinition('one', Int),
            InputDefinition('_two', Nothing),
            InputDefinition('two', Int),
            InputDefinition('_three', Nothing),
            InputDefinition('three', Int),
        ]
    )
    def adder(_context, one, two, three):
        assert one == 1
        assert two == 2
        assert three == 3
        return one + two + three

    pipeline = PipelineDefinition(
        name='input_test',
        solids=[emit_one, emit_two, emit_three, emit_nothing, adder],
        dependencies={
            SolidInstance('emit_nothing', '_one'): {},
            SolidInstance('emit_nothing', '_two'): {},
            SolidInstance('emit_nothing', '_three'): {},
            'adder': {
                '_one': DependencyDefinition('_one'),
                '_two': DependencyDefinition('_two'),
                '_three': DependencyDefinition('_three'),
                'one': DependencyDefinition('emit_one'),
                'two': DependencyDefinition('emit_two'),
                'three': DependencyDefinition('emit_three'),
            },
        },
    )
    result = execute_pipeline(pipeline)
    assert result.success



from dagster import check
from dagster.core.definitions import InputDefinition, PipelineDefinition, Solid, SolidHandle
from dagster.core.errors import DagsterInvariantViolationError

from .objects import (
    ExecutionStep,
    SingleOutputStepCreationData,
    StepKind,
    StepOutput,
    StepOutputValue,
)

INPUT_THUNK_OUTPUT = 'input_thunk_output'


def _create_input_thunk_execution_step(pipeline_def, solid, input_def, input_spec):
    check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition)
    check.inst_param(solid, 'solid', Solid)
    check.inst_param(input_def, 'input_def', InputDefinition)

    check.invariant(input_def.runtime_type.input_schema)

    def _fn(step_context, _inputs):
        value = input_def.runtime_type.input_schema.construct_from_config_value(
            step_context, input_spec
        )
        yield StepOutputValue(output_name=INPUT_THUNK_OUTPUT, value=value)

    return ExecutionStep(
        pipeline_name=pipeline_def.name,
        key_suffix='inputs.' + input_def.name + '.read',
        step_inputs=[],
        step_outputs=[
            StepOutput(name=INPUT_THUNK_OUTPUT, runtime_type=input_def.runtime_type, optional=False)
        ],
        compute_fn=_fn,
        kind=StepKind.INPUT_THUNK,
        solid_handle=SolidHandle(solid.name, solid.definition.name),
        logging_tags={'input': input_def.name},
    )


def create_input_thunk_execution_step(pipeline_def, solid, input_def, input_spec):
    check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition)
    check.inst_param(solid, 'solid', Solid)
    check.inst_param(input_def, 'input_def', InputDefinition)

    dependency_structure = pipeline_def.dependency_structure
    input_handle = solid.input_handle(input_def.name)

    if dependency_structure.has_dep(input_handle):
        raise DagsterInvariantViolationError(
            (
                'In pipeline {pipeline_name} solid {solid_name}, input {input_name} '
                'you have specified an input via config while also specifying '
                'a dependency. Either remove the dependency, specify a subdag '
                'to execute, or remove the inputs specification in the environment.'
            ).format(
                pipeline_name=pipeline_def.name, solid_name=solid.name, input_name=input_def.name
            )
        )

    input_thunk = _create_input_thunk_execution_step(pipeline_def, solid, input_def, input_spec)
    return SingleOutputStepCreationData(input_thunk, INPUT_THUNK_OUTPUT)



from collections import namedtuple

from dagster import check

from .utils import DEFAULT_OUTPUT


class Result(namedtuple('_Result', 'value output_name')):
    '''A solid transform function return a stream of Result objects.
    An implementator of a SolidDefinition must provide a transform that
    yields objects of this type.

    Attributes:
        value (Any): Value returned by the transform.
        output_name (str): Name of the output returns. defaults to "result"
'''

    def __new__(cls, value, output_name=DEFAULT_OUTPUT):
        return super(Result, cls).__new__(cls, value, check.str_param(output_name, 'output_name'))



from dagster.core import types


from dagster.core.execution.api import execute_pipeline, execute_pipeline_iterator

from dagster.core.execution.results import PipelineExecutionResult, SolidExecutionResult

from dagster.core.execution.context_creation_pipeline import PipelineConfigEvaluationError

from dagster.core.execution.init_context import InitContext, InitResourceContext

from dagster.core.execution.execution_context import (
    InProcessExecutorConfig,
    MultiprocessExecutorConfig,
    RunConfig,
)

from dagster.core.execution.user_context import ExecutionContext

from dagster.core.runs import RunStorageMode


from dagster.core.definitions import (
    DependencyDefinition,
    ExpectationDefinition,
    ExpectationResult,
    InputDefinition,
    OutputDefinition,
    PipelineContextDefinition,
    PipelineDefinition,
    RepositoryDefinition,
    Result,
    Materialization,
    ModeDefinition,
    SolidDefinition,
    SolidInstance,
)

from dagster.core.definitions.resource import ResourceDefinition, resource
from dagster.core.definitions.decorators import MultipleResults, lambda_solid, solid

from dagster.core.events import DagsterEventType

from dagster.core.errors import (
    DagsterExecutionStepExecutionError,
    DagsterExecutionStepNotFoundError,
    DagsterExpectationFailedError,
    DagsterInvalidDefinitionError,
    DagsterInvariantViolationError,
    DagsterRuntimeCoercionError,
    DagsterTypeError,
    DagsterUserCodeExecutionError,
    DagsterUserError,
    DagsterStepOutputNotFoundError,
)


from dagster.core.types import (
    Any,
    Bool,
    Dict,
    Field,
    Float,
    input_schema,
    input_selector_schema,
    Int,
    List,
    NamedDict,
    NamedSelector,
    Nullable,
    output_schema,
    output_selector_schema,
    Path,
    PermissiveDict,
    PythonObjectType,
    Selector,
    String,
    Nothing,
)

from dagster.core.types.decorator import dagster_type, as_dagster_type
from dagster.core.types.marshal import SerializationStrategy
from dagster.core.types.config import ConfigType, ConfigScalar, Enum, EnumValue
from dagster.core.types.evaluator import DagsterEvaluateConfigValueError
from dagster.core.types.runtime import Bytes, RuntimeType

from dagster.utils.test import execute_solid, execute_solids

from .version import __version__


__all__ = [
    # Definition
    'DependencyDefinition',
    'ExpectationDefinition',
    'ExpectationResult',
    'Field',
    'InputDefinition',
    'Materialization',
    'ModeDefinition',
    'OutputDefinition',
    'PipelineContextDefinition',
    'PipelineDefinition',
    'RepositoryDefinition',
    'ResourceDefinition',
    'Result',
    'SolidDefinition',
    'SolidInstance',
    # Decorators
    'lambda_solid',
    'resource',
    'solid',
    'MultipleResults',
    # Execution
    'execute_pipeline_iterator',
    'execute_pipeline',
    'DagsterEventType',
    'ExecutionContext',
    'InitContext',
    'InitResourceContext',
    'InProcessExecutorConfig',
    'MultiprocessExecutorConfig',
    'PipelineConfigEvaluationError',
    'PipelineExecutionResult',
    'RunConfig',
    'RunStorageMode',
    'SolidExecutionResult',
    # Errors
    'DagsterEvaluateConfigValueError',
    'DagsterExecutionStepExecutionError',
    'DagsterExecutionStepNotFoundError',
    'DagsterExpectationFailedError',
    'DagsterInvalidDefinitionError',
    'DagsterInvariantViolationError',
    'DagsterRuntimeCoercionError',
    'DagsterTypeError',
    'DagsterUserCodeExecutionError',
    'DagsterUserError',
    'DagsterStepOutputNotFoundError',
    # Utilities
    'execute_solid',
    'execute_solids',
    # types
    'Any',
    'Bool',
    'Bytes',
    'input_schema',
    'input_selector_schema',
    'Dict',
    'Enum',
    'EnumValue',
    'Float',
    'Int',
    'List',
    'NamedDict',
    'NamedSelector',
    'Nullable',
    'output_schema',
    'output_selector_schema',
    'Path',
    'PermissiveDict',
    'PythonObjectType',
    'Selector',
    'String',
    'SerializationStrategy',
    'Nothing',
    # type creation
    'as_dagster_type',
    'dagster_type',
    'RuntimeType',
    'ConfigType',
    'ConfigScalar',
]



from dagster import DependencyDefinition, InputDefinition, PipelineDefinition, lambda_solid


@lambda_solid
def solid_one():
    return 'foo'


@lambda_solid(inputs=[InputDefinition('arg_one')])
def solid_two(arg_one):
    return arg_one * 2


def define_hello_dag_pipeline():
    return PipelineDefinition(
        name='hello_dag_pipeline',
        solids=[solid_one, solid_two],
        dependencies={'solid_two': {'arg_one': DependencyDefinition('solid_one')}},
    )



from dagster.utils import safe_isfile, script_relative_path


def test_safe_isfile():
    assert safe_isfile(script_relative_path('test_safe_isfile.py'))
    assert not safe_isfile(script_relative_path('test_safe_isfile_foobar.py'))



import os
import shutil
import tempfile
import uuid

import pytest

from dagster import check, String, Nullable, seven, List, Bool
from dagster.core.types.marshal import SerializationStrategy
from dagster.core.object_store import FileSystemObjectStore, TypeStoragePlugin
from dagster.core.types.runtime import (
    Bool as RuntimeBool,
    resolve_to_runtime_type,
    RuntimeType,
    String as RuntimeString,
)
from dagster.utils import mkdir_p
from dagster.utils.test import yield_empty_pipeline_context


class UppercaseSerializationStrategy(SerializationStrategy):  # pylint: disable=no-init
    def serialize_value(self, _context, value, write_file_obj):
        return write_file_obj.write(bytes(value.upper().encode('utf-8')))

    def deserialize_value(self, _context, read_file_obj):
        return read_file_obj.read().decode('utf-8').lower()


class LowercaseString(RuntimeType):
    def __init__(self):
        super(LowercaseString, self).__init__(
            'lowercase_string',
            'LowercaseString',
            serialization_strategy=UppercaseSerializationStrategy(),
        )


class FancyStringFilesystemTypeStoragePlugin(TypeStoragePlugin):  # pylint:disable=no-init
    @classmethod
    def set_object(cls, object_store, obj, context, runtime_type, paths):
        check.inst_param(object_store, 'object_store', FileSystemObjectStore)
        paths.append(obj)
        mkdir_p(os.path.join(object_store.root, *paths))

    @classmethod
    def get_object(cls, object_store, context, runtime_type, paths):
        check.inst_param(object_store, 'object_store', FileSystemObjectStore)
        return os.listdir(os.path.join(object_store.root, *paths))[0]


def test_file_system_object_store():
    run_id = str(uuid.uuid4())

    object_store = FileSystemObjectStore(run_id=run_id)
    assert object_store.root == os.path.join(
        seven.get_system_temp_directory(), 'dagster', 'runs', run_id, 'files'
    )

    with yield_empty_pipeline_context(run_id=run_id) as context:
        try:
            object_store.set_object(True, context, RuntimeBool.inst(), ['true'])
            assert object_store.has_object(context, ['true'])
            assert object_store.get_object(context, RuntimeBool.inst(), ['true']) is True
            assert object_store.url_for_paths(['true']).startswith('file:///')
            assert object_store.rm_object(context, ['true']) is None
            assert object_store.rm_object(context, ['true']) is None
            assert object_store.rm_object(context, ['dslkfhjsdflkjfs']) is None
        finally:
            try:
                shutil.rmtree(object_store.root)
            except seven.FileNotFoundError:
                pass


def test_file_system_object_store_with_base_dir():
    run_id = str(uuid.uuid4())

    try:
        tempdir = tempfile.mkdtemp()

        object_store = FileSystemObjectStore(run_id=run_id, base_dir=tempdir)
        assert object_store.root == os.path.join(tempdir, 'dagster', 'runs', run_id, 'files')

        with yield_empty_pipeline_context(run_id=run_id) as context:
            try:
                object_store.set_object(True, context, RuntimeBool.inst(), ['true'])
                assert object_store.has_object(context, ['true'])
                assert object_store.get_object(context, RuntimeBool.inst(), ['true']) is True

            finally:
                try:
                    shutil.rmtree(object_store.root)
                except seven.FileNotFoundError:
                    pass
    finally:
        try:
            shutil.rmtree(tempdir)
        except seven.FileNotFoundError:
            pass


def test_file_system_object_store_composite_types():
    run_id = str(uuid.uuid4())

    object_store = FileSystemObjectStore(run_id=run_id)
    assert object_store.root == os.path.join(
        seven.get_system_temp_directory(), 'dagster', 'runs', run_id, 'files'
    )

    with yield_empty_pipeline_context(run_id=run_id) as context:
        try:
            object_store.set_object(
                [True, False], context, resolve_to_runtime_type(List(Bool)).inst(), ['bool']
            )
            assert object_store.has_object(context, ['bool'])
            assert object_store.get_object(
                context, resolve_to_runtime_type(List(Bool)).inst(), ['bool']
            ) == [True, False]

        finally:
            try:
                shutil.rmtree(object_store.root)
            except seven.FileNotFoundError:
                pass


def test_file_system_object_store_with_custom_serializer():
    run_id = str(uuid.uuid4())

    object_store = FileSystemObjectStore(run_id=run_id)

    with yield_empty_pipeline_context(run_id=run_id) as context:
        try:
            object_store.set_object('foo', context, LowercaseString.inst(), ['foo'])

            with open(os.path.join(object_store.root, 'foo'), 'rb') as fd:
                assert fd.read().decode('utf-8') == 'FOO'

            assert object_store.has_object(context, ['foo'])
            assert object_store.get_object(context, LowercaseString.inst(), ['foo']) == 'foo'
        finally:
            try:
                shutil.rmtree(object_store.root)
            except seven.FileNotFoundError:
                pass


def test_file_system_object_store_composite_types_with_custom_serializer_for_inner_type():
    run_id = str(uuid.uuid4())

    object_store = FileSystemObjectStore(run_id=run_id)
    assert object_store.root == os.path.join(
        seven.get_system_temp_directory(), 'dagster', 'runs', run_id, 'files'
    )

    with yield_empty_pipeline_context(run_id=run_id) as context:
        try:
            object_store.set_object(
                ['foo', 'bar'],
                context,
                resolve_to_runtime_type(List(LowercaseString)).inst(),
                ['list'],
            )
            assert object_store.has_object(context, ['list'])
            assert object_store.get_object(
                context, resolve_to_runtime_type(List(Bool)).inst(), ['list']
            ) == ['foo', 'bar']

        finally:
            try:
                shutil.rmtree(object_store.root)
            except seven.FileNotFoundError:
                pass


def test_file_system_object_store_with_type_storage_plugin():
    run_id = str(uuid.uuid4())

    # FIXME need a dedicated test bucket
    object_store = FileSystemObjectStore(
        run_id=run_id,
        types_to_register={RuntimeString.inst(): FancyStringFilesystemTypeStoragePlugin},
    )

    with yield_empty_pipeline_context(run_id=run_id) as context:
        try:
            object_store.set_value('hello', context, RuntimeString.inst(), ['obj_name'])

            assert object_store.has_object(context, ['obj_name'])
            assert object_store.get_value(context, RuntimeString.inst(), ['obj_name']) == 'hello'

        finally:
            object_store.rm_object(context, ['obj_name'])


def test_file_system_object_store_with_composite_type_storage_plugin():
    run_id = str(uuid.uuid4())

    # FIXME need a dedicated test bucket
    object_store = FileSystemObjectStore(
        run_id=run_id,
        types_to_register={RuntimeString.inst(): FancyStringFilesystemTypeStoragePlugin},
    )

    with yield_empty_pipeline_context(run_id=run_id) as context:
        with pytest.raises(check.NotImplementedCheckError):
            object_store.set_value(
                ['hello'], context, resolve_to_runtime_type(List(String)), ['obj_name']
            )

    with yield_empty_pipeline_context(run_id=run_id) as context:
        with pytest.raises(check.NotImplementedCheckError):
            object_store.set_value(
                ['hello'], context, resolve_to_runtime_type(Nullable(String)), ['obj_name']
            )

    with yield_empty_pipeline_context(run_id=run_id) as context:
        with pytest.raises(check.NotImplementedCheckError):
            object_store.set_value(
                ['hello'], context, resolve_to_runtime_type(List(Nullable(String))), ['obj_name']
            )

    with yield_empty_pipeline_context(run_id=run_id) as context:
        with pytest.raises(check.NotImplementedCheckError):
            object_store.set_value(
                ['hello'], context, resolve_to_runtime_type(Nullable(List(String))), ['obj_name']
            )
